#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2026, Rafal Kozlowski (@rkozlo) <rafalkozlowski07@gmail.com>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type

DOCUMENTATION = r'''
---
module: clickhouse_settings_profile

short_description: Creates, removes or modify a ClickHouse settings profile using the clickhouse-driver Client interface

description:
  - Creates, remove or modify a ClickHouse settings profile using the
    L(clickhouse-driver,https://clickhouse-driver.readthedocs.io/en/latest) Client interface.

attributes:
  check_mode:
    description: Supports check_mode.
    support: full
  idempotent:
    description:
      - Module can't determine what data type is certain setting.
        So when passing setting that will be resolved by server to another value module will be not idempotent.
        For example passing 1G as value server will store this as 100000000.
    support: partial

author:
  - Rafal Kozlowski (@rkozlo)

extends_documentation_fragment:
  - community.clickhouse.client_inst_opts
  - community.clickhouse.cluster_inst_opts
  - community.clickhouse.entity_inst_opts

version_added: '2.3.0'

options:
  state:
    description:
      - Whether the settings profile should be present or not.
      - If C(present), the module will create or update the settings profile to match the provided parameters.
      - If C(absent), the module will remove the settings profile if it exists.
    type: str
    choices: ["present", "absent"]
    default: "present"
  name:
    description:
      - The name of the row policy to manage.
    type: str
    required: true
'''

EXAMPLES = r'''
---
- name: Create plain settings profile
  community.clickhouse.clickhouse_settings_profile:
    name: test_profile

- name: Create settings profile with setting
  community.clickhouse.clickhouse_settings_profile:
    name: test_profile
    settings:
      max_threads:
        value: 1

- name: Create settings profile with settings
  community.clickhouse.clickhouse_settings_profile:
    name: test_profile
    settings:
      max_threads:
        value: 1
      max_memory_usage:
        value: 10000000000

- name: Create settings profile with profiles
  community.clickhouse.clickhouse_settings_profile:
    name: test_profile
    profiles:
      - web

- name: Create settings profile with settings and profiles
  community.clickhouse.clickhouse_settings_profile:
    name: test_profile
    settings:
      max_threads:
        value: 1
      max_memory_usage:
        value: 10000000000
    profiles:
      - web
- name: Drop settings profile
  community.clickhouse.clickhouse_settings_profile:
    name: test_profile
    state: absent
'''

RETURN = r'''
executed_statements:
  description:
  - Data-modifying executed statements.
  returned: on success
  type: list
  sample: [
    "CREATE SETTINGS PROFILE `test_profile` SETTINGS INHERIT `inh_1`, max_threads='1'",
    "ALTER SETTINGS PROFILE `test_profile` SETTINGS INHERIT `inh_1`, max_threads='1'",
    "DROP SETTINGS PROFILE `test_profile`",
  ]
'''

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.clickhouse.plugins.module_utils.clickhouse import (
    check_clickhouse_driver,
    client_common_argument_spec,
    connect_to_db_via_client,
    execute_query,
    get_main_conn_kwargs,
    get_on_cluster_clause,
    validate_identifier,
    cluster_argument_spec,
)
from ansible_collections.community.clickhouse.plugins.module_utils.entity_settings import (
    EntitySettings,
    get_settings_argument_spec,
)

executed_statements = []


class ClickHouseSettingsProfile:
    def __init__(self, module, client, name):
        self.module = module
        self.client = client
        validate_identifier(module, name, "collection name")
        self.name = name
        self.settings = EntitySettings(self.module, self.client, self.name, 'profile')
        self._exists = None

    @property
    def exists(self):
        if self._exists is None:
            query = f"SELECT 1 FROM system.settings_profiles WHERE name = '{self.name}'"
            result = execute_query(self.module, self.client, query)
            if result:
                self._exists = True
            else:
                self._exists = False
        return self._exists

    def create(self, settings, profiles, cluster=None):

        query = f"CREATE SETTINGS PROFILE `{self.name}`"
        query += get_on_cluster_clause(self.module, cluster)
        settings_entity = self.settings.compare_and_build_clause(settings, profiles)
        query += settings_entity[1]

        executed_statements.append(query)

        if not self.module.check_mode:
            execute_query(self.module, self.client, query)
        return True

    def alter(self, settings, profiles, cluster=None):
        if not self.exists:
            return False
        query = f"ALTER SETTINGS PROFILE `{self.name}`"
        query += get_on_cluster_clause(self.module, cluster)

        settings_entity = self.settings.compare_and_build_clause(settings, profiles)

        if settings_entity[0] is False:
            return False
        query += settings_entity[1]
        executed_statements.append(query)

        if not self.module.check_mode:
            execute_query(self.module, self.client, query)
        return True

    def drop(self, cluster=None):
        query = f"DROP SETTINGS PROFILE `{self.name}`"
        query += get_on_cluster_clause(self.module, cluster)
        executed_statements.append(query)

        if not self.module.check_mode:
            execute_query(self.module, self.client, query)
        return True


def main():
    # Set up arguments.
    # If there are common arguments shared across several modules,
    # create the common_argument_spec() function under plugins/module_utils/*
    # and invoke here to return a dict with those arguments
    argument_spec = client_common_argument_spec()
    argument_spec.update(
        state=dict(type="str", choices=["present", "absent"], default="present"),
        name=dict(type="str", required=True),
    )
    argument_spec.update(cluster_argument_spec())
    argument_spec.update(get_settings_argument_spec())

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
    )

    # Assign passed options to variables
    client_kwargs = module.params["client_kwargs"]
    # The reason why these arguments are separate from client_kwargs
    # is that we need to protect some sensitive data like passwords passed
    # to the module from logging (see the arguments above with no_log=True);
    # Such data must be passed as module arguments (not nested deep in values).
    main_conn_kwargs = get_main_conn_kwargs(module)
    state = module.params["state"]
    name = module.params["name"]
    cluster = module.params["cluster"]
    profiles = module.params["profiles"]
    settings = module.params["settings"]

    # Will fail if no driver informing the user
    check_clickhouse_driver(module)

    # Connect to DB
    client = connect_to_db_via_client(module, main_conn_kwargs, client_kwargs)

    # Do the job
    changed = False
    settings_profile = ClickHouseSettingsProfile(module, client, name)
    if state == "present":
        if not settings_profile.exists:
            changed = settings_profile.create(settings, profiles, cluster)
        else:
            changed = settings_profile.alter(settings, profiles, cluster)
    else:
        if settings_profile.exists:
            changed = settings_profile.drop(cluster)

    # Close connection
    client.disconnect_connection()

    # Users will get this in JSON output after execution
    module.exit_json(changed=changed, executed_statements=executed_statements)


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or
# https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
module: organizations_wireless_devices_power_mode_history_info
short_description: Information module for organizations _wireless _devices _power
  _mode _history
description:
  - Get all organizations _wireless _devices _power _mode _history. - > Return a record
    of power mode changes for wireless devices in the organization. For each device,
    it provides a series of events with timestamps indicating when a power mode change
    occurred and the new mode. The events are ordered by timestamp.
version_added: '1.0.0'
extends_documentation_fragment:
  - cisco.meraki.module_info
  - cisco.meraki.module_info_pagination
author: Francisco Munoz (@fmunoz)
options:
  headers:
    description: Additional headers.
    type: dict
  organizationId:
    description:
      - OrganizationId path parameter. Organization ID.
    type: str
  t0:
    description:
      - T0 query parameter. The beginning of the timespan for the data. The maximum
        lookback period is 1 day from today.
    type: str
  t1:
    description:
      - T1 query parameter. The end of the timespan for the data. T1 can be a maximum
        of 1 day after t0.
    type: str
  timespan:
    description:
      - >
        Timespan query parameter. The timespan for which the information will be fetched.
        If specifying timespan, do not specify parameters t0 and t1. The value must
        be in seconds and be less than or equal to 1 day. The default is 1 day.
    type: float
  perPage:
    description:
      - PerPage query parameter. The number of entries per page returned. Acceptable
        range is 3 - 20. Default is 10.
    type: int
  startingAfter:
    description:
      - >
        StartingAfter query parameter. A token used by the server to indicate the
        start of the page. Often this is a timestamp or an ID but it is not limited
        to those. This parameter should not be defined by client applications. The
        link for the first, last, prev, or next page in the HTTP Link header should
        define it.
    type: str
  endingBefore:
    description:
      - >
        EndingBefore query parameter. A token used by the server to indicate the end
        of the page. Often this is a timestamp or an ID but it is not limited to those.
        This parameter should not be defined by client applications. The link for
        the first, last, prev, or next page in the HTTP Link header should define
        it.
    type: str
  networkIds:
    description:
      - NetworkIds query parameter. Optional parameter to filter the result set by
        the included set of network IDs.
    elements: str
    type: list
  serials:
    description:
      - Serials query parameter. Optional parameter to filter device availabilities
        history by device serial numbers.
    elements: str
    type: list
requirements:
  - meraki >= 2.4.9
  - python >= 3.5
seealso:
  - name: Cisco Meraki documentation for wireless getOrganizationWirelessDevicesPowerModeHistory
    description: Complete reference of the getOrganizationWirelessDevicesPowerModeHistory
      API.
    link: https://developer.cisco.com/meraki/api-v1/#!get-organization-wireless-devices-power-mode-history
notes:
  - SDK Method used are
    wireless.Wireless.get_organization_wireless_devices_power_mode_history,
  - Paths used are
    get /organizations/{organizationId}/wireless/devices/power/mode/history,
"""

EXAMPLES = r"""
- name: Get all organizations _wireless _devices _power _mode _history
  cisco.meraki.organizations_wireless_devices_power_mode_history_info:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    t0: string
    t1: string
    timespan: 0
    perPage: 0
    startingAfter: string
    endingBefore: string
    networkIds: []
    serials: []
    organizationId: string
    total_pages: -1
    direction: next
  register: result
"""
RETURN = r"""
meraki_response:
  description: A dictionary or list with the response returned by the Cisco Meraki Python SDK
  returned: always
  type: list
  elements: dict
  sample: >
    [
      {
        "serial": "string",
        "model": "string",
        "name": "string",
        "mac": "string",
        "tags": [
          "string"
        ],
        "network": {
          "id": "string",
          "name": "string",
          "tags": [
            "string"
          ]
        },
        "events": [
          {
            "ts": "string",
            "powerMode": "string"
          }
        ]
      }
    ]
"""



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2026, Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


DOCUMENTATION = r'''
---
module: mysql_clone

short_description: Clone a MySQL instance from a donor server

description:
- Starts a MySQL remote clone operation on the recipient server and waits until the clone reaches a terminal state.
- Validates MySQL Clone plugin prerequisites on the recipient server but does not install or manage the plugin.
- The module is limited to MySQL Clone plugin behavior and fails on MariaDB.

author:
- Ron Gershburg (@rgershbu)

version_added: '5.1.0'

options:
  donor_host:
    description:
    - Hostname or IP address of the donor MySQL server.
    type: str
    required: true
  donor_port:
    description:
    - TCP port of the donor MySQL server.
    type: int
    default: 3306
  donor_user:
    description:
    - User account used by the recipient to connect to the donor for clone.
    type: str
    required: true
  donor_password:
    description:
    - Password used by the recipient to connect to the donor for clone.
    - The password is sent to MySQL as part of the C(CLONE INSTANCE FROM ...) statement.
    type: str
    required: true
  require_ssl:
    description:
    - Whether to append C(REQUIRE SSL) or C(REQUIRE NO SSL) to the clone statement.
    - If omitted, no SSL clause is added and MySQL server defaults apply.
    type: bool
  wait_timeout:
    description:
    - Maximum number of seconds to wait for the recipient to reconnect and the clone to reach a terminal state.
    type: int
    default: 1800
  poll_interval:
    description:
    - Number of seconds to wait between reconnect and status polling attempts.
    type: int
    default: 5

notes:
  - MySQL Clone plugin must already be installed and active on the recipient server.
  - Recipient-side clone variables such as C(clone_valid_donor_list) must be configured before running the module.
  - The module is not idempotent. Running it again starts a new clone operation if MySQL accepts it.
  - In check mode the module validates static prerequisites and reports that clone would be started, but does not execute clone.
  - MariaDB is not supported. Use backup and replication workflows instead.

attributes:
  check_mode:
    support: partial
    details:
      - In check mode the module does not start clone, but it still validates static recipient-side prerequisites.
  idempotent:
    support: partial
    details:
      - The module is not idempotent because each successful invocation starts a new clone operation.

seealso:
  - module: ansible.mysql.mysql_variables
  - module: ansible.mysql.mysql_replication

extends_documentation_fragment:
- ansible.mysql.mysql
'''

EXAMPLES = r'''
- name: Configure recipient donor allow-list separately
  ansible.mysql.mysql_variables:
    variable: clone_valid_donor_list
    value: "192.0.2.10:3306"
    mode: persist

- name: Start clone and wait for completion
  ansible.mysql.mysql_clone:
    donor_host: 192.0.2.10
    donor_port: 3306
    donor_user: clone_user
    donor_password: supersecret
    login_unix_socket: /run/mysqld/mysqld.sock

- name: Preview clone action in check mode
  check_mode: true
  ansible.mysql.mysql_clone:
    donor_host: 192.0.2.10
    donor_user: clone_user
    donor_password: supersecret
    login_unix_socket: /run/mysqld/mysqld.sock
'''

RETURN = r'''
query:
  description: Clone statement executed or predicted, with the password redacted.
  returned: always
  type: str
  sample: "CLONE INSTANCE FROM 'clone_user'@'192.0.2.10':3306 IDENTIFIED BY '********'"
clone_status:
  description: Final row from C(performance_schema.clone_status) for the current or last clone operation.
  returned: on success or on clone failure after status was collected
  type: dict
clone_progress:
  description: Rows from C(performance_schema.clone_progress) collected when the clone reaches a terminal state.
  returned: on success or on clone failure after status was collected
  type: list
  elements: dict
'''

import time

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.common.text.converters import to_native

from ansible_collections.ansible.mysql.plugins.module_utils.mysql import (
    get_server_implementation,
    get_server_version,
    mysql_common_argument_spec,
    mysql_connect,
    mysql_driver,
    mysql_driver_fail_msg,
)
from ansible_collections.ansible.mysql.plugins.module_utils.version import LooseVersion


PLUGIN_QUERY = (
    "SELECT PLUGIN_STATUS AS plugin_status "
    "FROM INFORMATION_SCHEMA.PLUGINS WHERE PLUGIN_NAME = 'clone'"
)
CLONE_STATUS_QUERY = "SELECT * FROM performance_schema.clone_status"
CLONE_PROGRESS_QUERY = "SELECT * FROM performance_schema.clone_progress"
TERMINAL_STATES = frozenset(('Completed', 'Failed'))


def validate_clone_support(module, server_implementation, server_version):
    if server_implementation != 'mysql':
        module.fail_json(msg='MariaDB is not supported by mysql_clone.')

    if LooseVersion(server_version) < LooseVersion('8.0.17'):
        module.fail_json(msg='mysql_clone requires MySQL 8.0.17 or newer.')


def build_clone_query(donor_host, donor_port, donor_user, donor_password, require_ssl=None):
    query = 'CLONE INSTANCE FROM %s@%s:%s IDENTIFIED BY %s'
    params = (donor_user, donor_host, donor_port, donor_password)

    if require_ssl is True:
        query += ' REQUIRE SSL'
    elif require_ssl is False:
        query += ' REQUIRE NO SSL'

    return query, params


def _quote_sql_value(value):
    if isinstance(value, bool):
        return '1' if value else '0'

    if isinstance(value, int):
        return str(value)

    return "'%s'" % str(value).replace("'", "''")


def get_redacted_query(query, params):
    redacted_params = list(params)
    if len(redacted_params) >= 4:
        redacted_params[3] = '********'

    pieces = query.split('%s')
    formatted = []
    for idx, piece in enumerate(pieces[:-1]):
        formatted.append(piece)
        formatted.append(_quote_sql_value(redacted_params[idx]))
    formatted.append(pieces[-1])
    return ''.join(formatted)


def is_terminal_state(state):
    return state in TERMINAL_STATES


def should_wait_after_execute_error(error_message):
    if error_message is None:
        return False

    if 'Restart server failed' in error_message:
        return True

    if error_message.startswith('(3707,'):
        return True

    return False


def _close_connection(cursor, connection):
    try:
        if cursor is not None:
            cursor.close()
    except Exception:
        pass

    try:
        if connection is not None:
            connection.close()
    except Exception:
        pass


def _connect(module):
    return mysql_connect(
        module,
        module.params['login_user'],
        module.params['login_password'],
        module.params['config_file'],
        module.params['client_cert'],
        module.params['client_key'],
        module.params['ca_cert'],
        'mysql',
        cursor_class='DictCursor',
        connect_timeout=module.params['connect_timeout'],
        check_hostname=module.params['check_hostname'],
    )


def ensure_clone_plugin_active(module, cursor):
    cursor.execute(PLUGIN_QUERY)
    result = cursor.fetchone()
    if not result or result.get('plugin_status') != 'ACTIVE':
        module.fail_json(msg='MySQL Clone plugin is not active on the recipient server.')


def validate_donor_allowed(module, cursor, donor_host, donor_port):
    cursor.execute("SHOW GLOBAL VARIABLES LIKE 'clone_valid_donor_list'")
    rows = cursor.fetchall()
    if not rows:
        module.fail_json(msg='clone_valid_donor_list is not available on the recipient server.')

    allowed_value = rows[0].get('Value') or ''
    allowed_donors = [item.strip() for item in allowed_value.split(',') if item.strip()]
    donor = '%s:%s' % (donor_host, donor_port)
    if donor not in allowed_donors:
        module.fail_json(msg='Recipient clone_valid_donor_list must contain %s.' % donor)


def fetch_clone_status(cursor):
    cursor.execute(CLONE_STATUS_QUERY)
    rows = cursor.fetchall()
    if not rows:
        return None
    return dict(rows[0])


def fetch_clone_progress(cursor):
    cursor.execute(CLONE_PROGRESS_QUERY)
    return [dict(row) for row in cursor.fetchall()]


def ensure_clone_not_running(module, status):
    if status and status.get('STATE') == 'In Progress':
        module.fail_json(msg='A clone operation is already in progress on the recipient server.')


def wait_for_clone_completion(module, execute_error=None):
    timeout_at = time.time() + module.params['wait_timeout']
    last_status = None
    last_progress = []
    last_error = None

    while time.time() <= timeout_at:
        cursor = None
        connection = None
        try:
            cursor, connection = _connect(module)
            last_status = fetch_clone_status(cursor)
            last_progress = fetch_clone_progress(cursor)
            if last_status and is_terminal_state(last_status.get('STATE')):
                return last_status, last_progress
        except Exception as exc:
            last_error = to_native(exc)
        finally:
            _close_connection(cursor, connection)

        time.sleep(module.params['poll_interval'])

    if last_status is not None:
        module.fail_json(
            msg='Timed out while waiting for clone to reach a terminal state.',
            clone_status=last_status,
            clone_progress=last_progress,
        )

    if last_error is not None:
        message = 'Timed out while reconnecting to the recipient server: %s' % last_error
        if execute_error:
            message += ' Original clone execution result: %s' % execute_error
        module.fail_json(msg=message)

    module.fail_json(msg='Timed out while waiting for clone status to become available.')


def main():
    argument_spec = mysql_common_argument_spec()
    argument_spec.update(
        donor_host=dict(type='str', required=True),
        donor_port=dict(type='int', default=3306),
        donor_user=dict(type='str', required=True),
        donor_password=dict(type='str', required=True, no_log=True),
        require_ssl=dict(type='bool', default=None),
        wait_timeout=dict(type='int', default=1800),
        poll_interval=dict(type='int', default=5),
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
    )

    if module.params['donor_port'] < 1 or module.params['donor_port'] > 65535:
        module.fail_json(msg='donor_port must be a valid unix port number (1-65535)')

    if module.params['wait_timeout'] <= 0:
        module.fail_json(msg='wait_timeout must be greater than 0')

    if module.params['poll_interval'] <= 0:
        module.fail_json(msg='poll_interval must be greater than 0')

    if mysql_driver is None:
        module.fail_json(msg=mysql_driver_fail_msg)

    query, params = build_clone_query(
        module.params['donor_host'],
        module.params['donor_port'],
        module.params['donor_user'],
        module.params['donor_password'],
        module.params['require_ssl'],
    )
    redacted_query = get_redacted_query(query, params)

    cursor = None
    connection = None
    try:
        cursor, connection = _connect(module)
        server_implementation = get_server_implementation(cursor)
        server_version = get_server_version(cursor)
        validate_clone_support(module, server_implementation, server_version)
        ensure_clone_plugin_active(module, cursor)
        validate_donor_allowed(module, cursor, module.params['donor_host'], module.params['donor_port'])
        ensure_clone_not_running(module, fetch_clone_status(cursor))

        if module.check_mode:
            module.exit_json(
                changed=True,
                query=redacted_query,
                msg='Clone would be started.',
            )

        execute_error = None
        try:
            cursor.execute(query, params)
        except Exception as exc:
            execute_error = to_native(exc)
            if not should_wait_after_execute_error(execute_error):
                module.fail_json(
                    msg='Clone failed to start. %s' % execute_error,
                    query=redacted_query,
                )
    finally:
        _close_connection(cursor, connection)

    status, progress = wait_for_clone_completion(module, execute_error=execute_error)
    state = status.get('STATE')

    if state == 'Completed':
        module.exit_json(
            changed=True,
            msg='Clone completed successfully.',
            query=redacted_query,
            clone_status=status,
            clone_progress=progress,
        )

    message = 'Clone failed.'
    if status.get('ERROR_MESSAGE'):
        message = '%s %s' % (message, status['ERROR_MESSAGE'])
    elif execute_error:
        message = '%s %s' % (message, execute_error)

    module.fail_json(
        msg=message.strip(),
        query=redacted_query,
        clone_status=status,
        clone_progress=progress,
    )


if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2026, Ansible community
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

DOCUMENTATION = r'''
---
module: mysql_resource_group_info

short_description: Gather information about MySQL resource groups

description:
  - Gather information about MySQL resource groups.
  - Resource groups are supported by MySQL 8.0 or later.

version_added: '5.1.0'

options:
  name:
    description:
      - Limit the collected information to a single resource group name.
    type: str

notes:
  - This module supports MySQL 8.0 or later only.
  - Resource groups are not available on MariaDB.

attributes:
  check_mode:
    support: full
  idempotent:
    support: full

seealso:
  - module: ansible.mysql.mysql_resource_group
  - name: INFORMATION_SCHEMA RESOURCE_GROUPS reference
    description: Complete reference of the INFORMATION_SCHEMA RESOURCE_GROUPS table documentation.
    link: https://dev.mysql.com/doc/refman/8.4/en/information-schema-resource-groups-table.html

author:
  - Ron Gershburg (@ronger4)

extends_documentation_fragment:
  - ansible.mysql.mysql
'''

EXAMPLES = r'''
# If you encounter the "Please explicitly state intended protocol" error,
# use the login_unix_socket argument
- name: Gather all resource groups
  ansible.mysql.mysql_resource_group_info:
    login_unix_socket: /run/mysqld/mysqld.sock

- name: Gather a single resource group
  ansible.mysql.mysql_resource_group_info:
    name: reporting
'''

RETURN = r'''
resource_groups:
  description: List of normalized resource groups.
  returned: always
  type: list
  elements: dict
  contains:
    name:
      description: Resource group name.
      type: str
      sample: reporting
    resource_group_type:
      description: Resource group type.
      type: str
      sample: USER
    enabled:
      description: Whether the resource group is enabled.
      type: bool
      sample: true
    vcpu_ids:
      description: Normalized CPU affinity.
      type: str
      sample: 4-7
    thread_priority:
      description: Resource-group thread priority.
      type: int
      sample: 5
'''

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.common.text.converters import to_native
from ansible_collections.ansible.mysql.plugins.module_utils.mysql import (
    mysql_common_argument_spec,
    mysql_connect,
    mysql_driver,
    mysql_driver_fail_msg,
)
from ansible_collections.ansible.mysql.plugins.module_utils.resource_group import (
    ensure_resource_groups_supported,
    normalize_resource_group_row,
)


def get_resource_groups_info(cursor, name=None):
    query = (
        'SELECT RESOURCE_GROUP_NAME, RESOURCE_GROUP_TYPE, RESOURCE_GROUP_ENABLED, VCPU_IDS, THREAD_PRIORITY '
        'FROM INFORMATION_SCHEMA.RESOURCE_GROUPS'
    )
    params = None

    if name:
        query += ' WHERE RESOURCE_GROUP_NAME = %s'
        params = (name,)

    query += ' ORDER BY RESOURCE_GROUP_NAME'

    if params:
        cursor.execute(query, params)
    else:
        cursor.execute(query)

    return {
        'resource_groups': [normalize_resource_group_row(row) for row in cursor.fetchall()]
    }


def main():
    argument_spec = mysql_common_argument_spec()
    argument_spec.update(
        name=dict(type='str'),
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
    )

    if mysql_driver is None:
        module.fail_json(msg=mysql_driver_fail_msg)

    login_user = module.params['login_user']
    login_password = module.params['login_password']
    config_file = module.params['config_file']
    ssl_cert = module.params['client_cert']
    ssl_key = module.params['client_key']
    ssl_ca = module.params['ca_cert']
    connect_timeout = module.params['connect_timeout']
    check_hostname = module.params['check_hostname']
    name = module.params['name']

    try:
        cursor, db_conn = mysql_connect(
            module,
            login_user,
            login_password,
            config_file,
            ssl_cert,
            ssl_key,
            ssl_ca,
            connect_timeout=connect_timeout,
            check_hostname=check_hostname,
            cursor_class='DictCursor',
        )
    except Exception as e:
        module.fail_json(msg='unable to connect to database: %s' % to_native(e))

    ensure_resource_groups_supported(module, cursor)
    module.exit_json(changed=False, **get_resource_groups_info(cursor, name))


if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}

DOCUMENTATION = r'''
---
module: intersight_link_control_policy
short_description: Link Control Policy configuration for Cisco Intersight
description:
  - Manages Link Control Policy configuration on Cisco Intersight.
  - A policy to configure UDLD (UniDirectional Link Detection) settings on Cisco Intersight managed devices.
  - For more information see L(Cisco Intersight,https://intersight.com/apidocs/fabric/LinkControlPolicy/get/).
extends_documentation_fragment: intersight
options:
  state:
    description:
      - If C(present), will verify the resource is present and will create if needed.
      - If C(absent), will verify the resource is absent and will delete if needed.
    type: str
    choices: [present, absent]
    default: present
  organization:
    description:
      - The name of the Organization this resource is assigned to.
      - Profiles, Policies, and Pools that are created within a Custom Organization are applicable only to devices in the same Organization.
    type: str
    default: default
  name:
    description:
      - The name assigned to the Link Control Policy.
      - The name must be between 1 and 64 alphanumeric characters, allowing special characters :-_.
    type: str
    required: true
  description:
    description:
      - The user-defined description for the Link Control Policy.
      - Description can contain letters(a-z, A-Z), numbers(0-9), hyphen(-), period(.), colon(:), or an underscore(_).
    type: str
    aliases: [descr]
  tags:
    description:
      - List of tags in Key:<user-defined key> Value:<user-defined value> format.
    type: list
    elements: dict
  udld_admin_state:
    description:
      - UDLD Admin configured Link State for this port.
    type: str
    choices: ['Enabled', 'Disabled']
    default: 'Enabled'
  udld_mode:
    description:
      - UDLD Admin configured Mode for this port.
      - Cannot be set to 'aggressive' when udld_admin_state is 'Disabled'.
    type: str
    choices: ['normal', 'aggressive']
    default: 'normal'
author:
  - Ron Gershburg (@rgershbu)
'''

EXAMPLES = r'''
- name: Configure Link Control Policy with UDLD enabled in aggressive mode
  cisco.intersight.intersight_link_control_policy:
    api_private_key: "{{ api_private_key }}"
    api_key_id: "{{ api_key_id }}"
    name: "test-link-aggressive"
    description: "Link Control Policy with aggressive UDLD"
    udld_admin_state: "Enabled"
    udld_mode: "aggressive"
    tags:
      - Key: Site
        Value: DataCenter-A

- name: Configure Link Control Policy with UDLD disabled
  cisco.intersight.intersight_link_control_policy:
    api_private_key: "{{ api_private_key }}"
    api_key_id: "{{ api_key_id }}"
    name: "test-link-disabled"
    description: "Link Control Policy with UDLD disabled"
    udld_admin_state: "Disabled"
    udld_mode: "normal"

- name: Delete Link Control Policy
  cisco.intersight.intersight_link_control_policy:
    api_private_key: "{{ api_private_key }}"
    api_key_id: "{{ api_key_id }}"
    name: "test-link-policy"
    state: absent
'''

RETURN = r'''
api_response:
  description: The API response output returned by the specified resource.
  returned: always
  type: dict
  sample:
    "api_response": {
        "Name": "test-link-policy",
        "ObjectType": "fabric.LinkControlPolicy",
        "UdldSettings": {
            "AdminState": "Enabled",
            "Mode": "aggressive",
            "ClassId": "fabric.UdldSettings",
            "ObjectType": "fabric.UdldSettings"
        },
        "Tags": [
            {
                "Key": "Site",
                "Value": "DataCenter-A"
            }
        ]
    }
'''


from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.intersight.plugins.module_utils.intersight import IntersightModule, intersight_argument_spec


def validate_udld_configuration(module):
    """
    Validate UDLD configuration to ensure aggressive mode is not used when admin state is disabled.
    """
    udld_admin_state = module.params['udld_admin_state']
    udld_mode = module.params['udld_mode']

    if udld_admin_state == 'Disabled' and udld_mode == 'aggressive':
        module.fail_json(
            msg="Cannot configure the link control policy. Mode should not be 'aggressive' when the policy is disabled."
        )


def main():
    argument_spec = intersight_argument_spec.copy()
    argument_spec.update(
        state=dict(type='str', choices=['present', 'absent'], default='present'),
        organization=dict(type='str', default='default'),
        name=dict(type='str', required=True),
        description=dict(type='str', aliases=['descr']),
        tags=dict(type='list', elements='dict'),
        udld_admin_state=dict(type='str', choices=['Enabled', 'Disabled'], default='Enabled'),
        udld_mode=dict(type='str', choices=['normal', 'aggressive'], default='normal')
    )
    module = AnsibleModule(
        argument_spec,
        supports_check_mode=True,
    )

    # Validate UDLD configuration
    validate_udld_configuration(module)

    intersight = IntersightModule(module)
    intersight.result['api_response'] = {}
    intersight.result['trace_id'] = ''

    # Resource path used to configure policy
    resource_path = '/fabric/LinkControlPolicies'
    # Define API body used in compares or create
    intersight.api_body = {
        'Organization': {
            'Name': intersight.module.params['organization'],
        },
        'Name': intersight.module.params['name']
    }

    if intersight.module.params['state'] == 'present':
        intersight.set_tags_and_description()
        intersight.api_body['UdldSettings'] = {
            'AdminState': intersight.module.params['udld_admin_state'],
            'Mode': intersight.module.params['udld_mode']
        }

    intersight.configure_policy_or_profile(resource_path=resource_path)

    module.exit_json(**intersight.result)


if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2023, Gaspard Micol (@gmicol) <gmicol@cisco.com>
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type

ANSIBLE_METADATA = {"metadata_version": "1.1", "status": ["preview"], "supported_by": "certified"}

DOCUMENTATION = r"""
---
module: aci_bgp_timers_policy
short_description: Manage BGP timers policy (bgp:CtxPol)
description:
- Manage BGP timers policies for the Tenants on Cisco ACI fabrics.
options:
  tenant:
    description:
    - The name of an existing tenant.
    type: str
    aliases: [ tenant_name ]
  bgp_timers_policy:
    description:
    - The name of the BGP timers policy.
    type: str
    aliases: [ bgp_timers_policy_name, name ]
  graceful_restart_controls:
    description:
    - The property to determine whether the entity functions only as a graceful restart helper or whether graceful restart is enabled completely.
    - The graceful restart helper option configures the local BGP router to support the graceful restart of a remote BGP peer.
    - The complete graceful restart option allows BGP graceful restart to be enabled or disable for an individual neighbor.
    - The APIC defaults to C(helper) when unset during creation.
    type: str
    choices: [ helper, complete ]
  hold_interval:
    description:
    - The time period to wait before declaring the neighbor device down.
    - The APIC defaults to C(180) when unset during creation.
    type: int
  keepalive_interval:
    description:
    - The interval time between sending keepalive messages.
    - The APIC defaults to C(60) when unset during creation.
    type: int
  max_as_limit:
    description:
    - The maximum AS limit, to discard routes that have excessive AS numbers.
    - The APIC defaults to C(0) when unset during creation.
    type: int
  stale_interval:
    description:
    - The maximum time that BGP keeps stale routes from the restarting BGP peer.
    - The APIC defaults to C(default) which is equal to 300 sec when unset during creation.
    type: int
  description:
    description:
    - Description for the BGP timers policy.
    type: str
    aliases: [ descr ]
  state:
    description:
    - Use C(present) or C(absent) for adding or removing.
    - Use C(query) for listing an object or multiple objects.
    type: str
    choices: [ absent, present, query ]
    default: present
  name_alias:
    description:
    - The alias for the current object. This relates to the nameAlias field in ACI.
    type: str
extends_documentation_fragment:
- cisco.aci.aci
- cisco.aci.annotation
- cisco.aci.owner

notes:
- The C(tenant) used must exist before using this module in your playbook.
  The M(cisco.aci.aci_tenant) module can be used for this.
seealso:
- module: cisco.aci.aci_tenant
- name: APIC Management Information Model reference
  description: More information about the internal APIC class B(bgp:CtxPol).
  link: https://developer.cisco.com/docs/apic-mim-ref/
author:
- Gaspard Micol (@gmicol)
"""

EXAMPLES = r"""
- name: Create a BGP timers policy
  cisco.aci.aci_bgp_timers_policy:
    host: apic
    username: admin
    password: SomeSecretPassword
    bgp_timers_policy: my_bgp_timers_policy
    graceful_restart_controls: complete
    hold_interval: 360
    keepalive_interval: 120
    max_as_limit: 1
    stale_interval: 600
    tenant: production
    state: present
  delegate_to: localhost

- name: Delete a BGP timers policy
  cisco.aci.aci_bgp_timers_policy:
    host: apic
    username: admin
    password: SomeSecretPassword
    bgp_timers_policy: my_bgp_timers_policy
    tenant: production
    state: absent
  delegate_to: localhost

- name: Query all BGP timers policies
  cisco.aci.aci_bgp_timers_policy:
    host: apic
    username: admin
    password: SomeSecretPassword
    state: query
  delegate_to: localhost
  register: query_result

- name: Query a specific BGP timers policy
  cisco.aci.aci_bgp_timers_policy:
    host: apic
    username: admin
    password: SomeSecretPassword
    bgp_timers_policy: my_bgp_timers_policy
    tenant: production
    state: query
  delegate_to: localhost
  register: query_result
"""

RETURN = r"""
current:
  description: The existing configuration from the APIC after the module has finished
  returned: success
  type: list
  sample:
    [
        {
            "fvTenant": {
                "attributes": {
                    "descr": "Production environment",
                    "dn": "uni/tn-production",
                    "name": "production",
                    "nameAlias": "",
                    "ownerKey": "",
                    "ownerTag": ""
                }
            }
        }
    ]
error:
  description: The error information as returned from the APIC
  returned: failure
  type: dict
  sample:
    {
        "code": "122",
        "text": "unknown managed object class foo"
    }
raw:
  description: The raw output returned by the APIC REST API (xml or json)
  returned: parse error
  type: str
  sample: '<?xml version="1.0" encoding="UTF-8"?><imdata totalCount="1"><error code="122" text="unknown managed object class foo"/></imdata>'
sent:
  description: The actual/minimal configuration pushed to the APIC
  returned: info
  type: list
  sample:
    {
        "fvTenant": {
            "attributes": {
                "descr": "Production environment"
            }
        }
    }
previous:
  description: The original configuration from the APIC before the module has started
  returned: info
  type: list
  sample:
    [
        {
            "fvTenant": {
                "attributes": {
                    "descr": "Production",
                    "dn": "uni/tn-production",
                    "name": "production",
                    "nameAlias": "",
                    "ownerKey": "",
                    "ownerTag": ""
                }
            }
        }
    ]
proposed:
  description: The assembled configuration from the user-provided parameters
  returned: info
  type: dict
  sample:
    {
        "fvTenant": {
            "attributes": {
                "descr": "Production environment",
                "name": "production"
            }
        }
    }
filter_string:
  description: The filter string used for the request
  returned: failure or debug
  type: str
  sample: ?rsp-prop-include=config-only
method:
  description: The HTTP method used for the request to the APIC
  returned: failure or debug
  type: str
  sample: POST
response:
  description: The HTTP response from the APIC
  returned: failure or debug
  type: str
  sample: OK (30 bytes)
status:
  description: The HTTP status from the APIC
  returned: failure or debug
  type: int
  sample: 200
url:
  description: The HTTP url used for the request to the APIC
  returned: failure or debug
  type: str
  sample: https://10.11.12.13/api/mo/uni/tn-production.json
"""

from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.aci.plugins.module_utils.aci import ACIModule, aci_argument_spec, aci_annotation_spec, aci_owner_spec
from ansible_collections.cisco.aci.plugins.module_utils.constants import MATCH_GRACEFUL_RESTART_CONTROLS_MAPPING


def main():
    argument_spec = aci_argument_spec()
    argument_spec.update(aci_annotation_spec())
    argument_spec.update(aci_owner_spec())
    argument_spec.update(
        tenant=dict(type="str", aliases=["tenant_name"]),  # Not required for querying all objects
        bgp_timers_policy=dict(type="str", aliases=["bgp_timers_policy_name", "name"]),  # Not required for querying all objects
        graceful_restart_controls=dict(type="str", choices=["helper", "complete"]),
        hold_interval=dict(type="int"),
        keepalive_interval=dict(type="int"),
        max_as_limit=dict(type="int"),
        stale_interval=dict(type="int"),
        description=dict(type="str", aliases=["descr"]),
        state=dict(type="str", default="present", choices=["absent", "present", "query"]),
        name_alias=dict(type="str"),
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_if=[
            ["state", "absent", ["bgp_timers_policy", "tenant"]],
            ["state", "present", ["bgp_timers_policy", "tenant"]],
        ],
    )

    bgp_timers_policy = module.params.get("bgp_timers_policy")
    graceful_restart_controls = MATCH_GRACEFUL_RESTART_CONTROLS_MAPPING.get(module.params.get("graceful_restart_controls"))
    hold_interval = module.params.get("hold_interval")
    keepalive_interval = module.params.get("keepalive_interval")
    max_as_limit = module.params.get("max_as_limit")
    stale_interval = module.params.get("stale_interval")
    description = module.params.get("description")
    state = module.params.get("state")
    tenant = module.params.get("tenant")
    name_alias = module.params.get("name_alias")

    aci = ACIModule(module)

    aci.construct_url(
        root_class=dict(
            aci_class="fvTenant",
            aci_rn="tn-{0}".format(tenant),
            module_object=tenant,
            target_filter={"name": tenant},
        ),
        subclass_1=dict(
            aci_class="bgpCtxPol",
            aci_rn="bgpCtxP-{0}".format(bgp_timers_policy),
            module_object=bgp_timers_policy,
            target_filter={"name": bgp_timers_policy},
        ),
    )

    aci.get_existing()

    if state == "present":
        aci.payload(
            aci_class="bgpCtxPol",
            class_config=dict(
                name=bgp_timers_policy,
                grCtrl=graceful_restart_controls,
                holdIntvl=hold_interval,
                kaIntvl=keepalive_interval,
                maxAsLimit=max_as_limit,
                staleIntvl=stale_interval,
                descr=description,
                nameAlias=name_alias,
            ),
        )

        aci.get_diff(aci_class="bgpCtxPol")

        aci.post_config()

    elif state == "absent":
        aci.delete_config()

    aci.exit_json()


if __name__ == "__main__":
    main()



#!/usr/bin/python
#
# Copyright (c) 2021 Aparna Patil(@techcon65)
#
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

DOCUMENTATION = '''
---
module: azure_rm_ipgroup_info

version_added: "1.6.0"

short_description: Get IP group facts

description:
    - Get facts for specified IP group or all IP groups in a given resource group.

options:
    resource_group:
        description:
            - Name of the resource group.
        type: str
    name:
        description:
            - Name of the IP group.
        type: str
    tags:
        description:
            -  Limit results by providing a list of tags. Format tags as 'key' or 'key:value'.
        type: list
        elements: str

extends_documentation_fragment:
    - azure.azcollection.azure

author:
    - Aparna Patil (@techcon65)

'''

EXAMPLES = '''
- name: Get facts for one IP group
  azure_rm_ipgroup_info:
    resource_group: myAzureResourceGroup
    name: myipgroup

- name: Get facts for all IP groups in resource group
  azure_rm_ipgroup_info:
    resource_group: myAzureResourceGroup
'''

RETURN = '''
ipgroups:
    description:
        - Gets a list of IP groups.
    returned: always
    type: list
    elements: dict
    sample: [
        {
            "etag": "c67388ea-6dab-481b-9387-bd441c0d32f8",
            "firewalls": [],
            "id": "/subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroups/MyAzureResourceGroup/providers/
                   Microsoft.Network/ipGroups/myipgroup",
            "ip_addresses": [
                "13.64.39.16/32",
                "40.74.146.80/31",
                "40.74.147.32/28"
            ],
            "location": "eastus",
            "name": "myipgroup",
            "provisioning_state": "Succeeded",
            "tags": {
                "key1": "value1"
            }
        }
    ]
'''

from ansible_collections.azure.azcollection.plugins.module_utils.azure_rm_common import AzureRMModuleBase

try:
    from azure.core.exceptions import ResourceNotFoundError
except Exception:
    # This is handled in azure_rm_common
    pass

AZURE_OBJECT_CLASS = 'IpGroup'


class AzureRMIPGroupInfo(AzureRMModuleBase):

    def __init__(self):

        # define user inputs into argument
        self.module_arg_spec = dict(
            name=dict(type='str'),
            resource_group=dict(type='str'),
            tags=dict(type='list', elements='str')
        )

        # store the results of the module operation
        self.results = dict(
            changed=False
        )

        self.name = None
        self.resource_group = None
        self.tags = None

        super(AzureRMIPGroupInfo, self).__init__(self.module_arg_spec, supports_check_mode=True, supports_tags=False, facts_module=True)

    def exec_module(self, **kwargs):

        for key in self.module_arg_spec:
            setattr(self, key, kwargs[key])

        results = []
        # list the conditions and results to return based on user input
        if self.name is not None:
            # if there is IP group name provided, return facts about that specific IP group
            results = self.get_item()
        elif self.resource_group:
            # all the IP groups listed in specific resource group
            results = self.list_resource_group()
        else:
            # all the IP groups in a subscription
            results = self.list_items()

        self.results['ipgroups'] = self.curated_items(results)

        return self.results

    def get_item(self):
        self.log('Get properties for {0}'.format(self.name))
        item = None
        results = []
        # get specific IP group
        try:
            item = self.network_client.ip_groups.get(self.resource_group, self.name)
        except ResourceNotFoundError:
            pass

        # serialize result
        if item and self.has_tags(item.tags, self.tags):
            results = [item]
        return results

    def list_resource_group(self):
        self.log('List all IP groups for resource group - {0}'.format(self.resource_group))
        try:
            response = self.network_client.ip_groups.list_by_resource_group(self.resource_group)
        except ResourceNotFoundError as exc:
            self.fail("Failed to list for resource group {0} - {1}".format(self.resource_group, str(exc)))

        results = []
        for item in response:
            if self.has_tags(item.tags, self.tags):
                results.append(item)
        return results

    def list_items(self):
        self.log('List all IP groups for a subscription ')
        try:
            response = self.network_client.ip_groups.list()
        except ResourceNotFoundError as exc:
            self.fail("Failed to list all items - {0}".format(str(exc)))

        results = []
        for item in response:
            if self.has_tags(item.tags, self.tags):
                results.append(item)
        return results

    def curated_items(self, raws):
        return [self.ipgroup_to_dict(item) for item in raws] if raws else []

    def ipgroup_to_dict(self, ipgroup):
        result = dict(
            id=ipgroup.id,
            name=ipgroup.name,
            location=ipgroup.location,
            tags=ipgroup.tags,
            ip_addresses=ipgroup.ip_addresses,
            provisioning_state=ipgroup.provisioning_state,
            firewalls=[dict(id=x.id) for x in ipgroup.firewalls],
            etag=ipgroup.etag
        )
        return result


def main():
    AzureRMIPGroupInfo()


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

# Copyright: Contributors to the Ansible project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

"""Generic text processing utilities for AWS plugins."""

from __future__ import annotations

import re

from ansible.module_utils.common.text.converters import to_text


def filter_ansi(line: str, is_windows: bool) -> str:
    """Remove any ANSI terminal control codes and PTY artifacts.

    :param line: The input line.
    :param is_windows: Whether the output is coming from a Windows host.
    :returns: The result line.
    """
    line = to_text(line)

    # Replace PTY line wrap sequences and normalise line endings
    line = line.replace("\r\r\n", "\n")
    line = line.replace("\r\n", "\n")
    # Remove standalone carriage returns (PTY artifacts from line wrapping)
    line = line.replace("\r", "")

    if is_windows:
        osc_filter = re.compile(r"\x1b\][^\x07]*\x07")
        line = osc_filter.sub("", line)
        ansi_filter = re.compile(r"(\x9B|\x1B\[)[0-?]*[ -/]*[@-~]")
        line = ansi_filter.sub("", line)

        if len(line) == 201:
            line = line[:-1]

    return line



# -*- coding: utf-8 -*-

# Copyright: Contributors to the Ansible project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

import typing

if typing.TYPE_CHECKING:
    from typing import Any
    from typing import Dict
    from typing import Optional
    from typing import Tuple

from ansible_collections.amazon.aws.plugins.plugin_utils.s3 import generate_encryption_settings


class S3ClientManager:
    def __init__(self, client: Any) -> None:
        """
        Initialise the S3ClientManager.

        :param client: The boto3 S3 client
        """
        self._s3_client = client

    @property
    def client(self) -> Any:
        return self._s3_client

    def get_url(
        self,
        client_method: str,
        bucket_name: str,
        out_path: str,
        http_method: str,
        extra_args: Optional[Dict[str, Any]] = None,
    ) -> str:
        """Generate URL for get_object / put_object"""

        params = {"Bucket": bucket_name, "Key": out_path}
        if extra_args is not None:
            params.update(extra_args)
        return self._s3_client.generate_presigned_url(
            client_method, Params=params, ExpiresIn=3600, HttpMethod=http_method
        )

    def generate_host_commands(
        self,
        bucket_name: str,
        bucket_sse_mode: Optional[str],
        bucket_sse_kms_key_id: Optional[str],
        s3_path: str,
        in_path: str,
        out_path: str,
        is_windows: bool,
        method: str,
    ) -> Tuple[str, Optional[Dict[str, str]]]:
        """
        Generate commands for the specified bucket and configurations, S3 path, input path, and output path.
        The function generates a curl/Invoke-WebRequest command to be executed on the managed node.
        The method 'get' means the controller node wants to retrieve content from the managed node, the corresponding
        on the managed node is a 'put' command to updload content to the S3 bucket.
        The method 'put' means the controller node wants to upload content to the managed node, the corresponding
        on the managed node is a 'get' command to download content to the S3 bucket.

        :param bucket_name: The name of the S3 bucket used for file transfers.
        :param s3_path: The S3 path to the file to be sent.
        :param in_path: Input path
        :param out_path: Output path
        :param method: The request method to use for the command (can be "get" or "put").

        :returns: A tuple containing a shell command string and optional extra arguments for the PUT request.
        """

        command = None
        url = None
        put_args = None
        if method == "get":
            put_args, put_headers = generate_encryption_settings(bucket_sse_mode, bucket_sse_kms_key_id)
            url = self.get_url("put_object", bucket_name, s3_path, "PUT", extra_args=put_args)
            if is_windows:
                put_command_headers = "; ".join([f"'{h}' = '{v}'" for h, v in put_headers.items()])
                # Use -Body with ReadAllBytes instead of -InFile to properly handle unicode paths
                # PowerShell's -InFile parameter can have issues with unicode file paths
                # Using .NET ReadAllBytes ensures proper unicode path handling
                escaped_in_path = in_path.replace("'", "''")
                command = (
                    "$ProgressPreference = 'SilentlyContinue' ; "
                    "$ErrorActionPreference = 'Stop' ; "
                    f"$fileContent = [System.IO.File]::ReadAllBytes('{escaped_in_path}') ; "
                    "Invoke-WebRequest -Method PUT "
                    # @{'key' = 'value'; 'key2' = 'value2'}
                    f"-Headers @{{{put_command_headers}}} "
                    "-Body $fileContent "
                    f"-Uri '{url}' "
                    f"-UseBasicParsing"
                )  # fmt: skip
            else:
                # Due to https://github.com/curl/curl/issues/183 earlier
                # versions of curl did not create the output file, when the
                # response was empty. Although this issue was fixed in 2015,
                # some actively maintained operating systems still use older
                # versions of it (e.g. CentOS 7)
                put_command_headers = " ".join([f"-H '{h}: {v}'" for h, v in put_headers.items()])
                command = (
                    "curl --request PUT "
                    f"{put_command_headers} "
                    f"--upload-file '{in_path}' "
                    f"'{url}'"
                )  # fmt: skip
        elif method == "put":
            url = self.get_url("get_object", bucket_name, s3_path, "GET")
            if is_windows:
                # Use .NET File.WriteAllBytes instead of -OutFile to properly handle unicode paths
                # PowerShell's -OutFile parameter can have issues with unicode file paths
                # Using .NET WriteAllBytes ensures proper unicode path handling
                escaped_out_path = out_path.replace("'", "''")
                command = (
                    "$ProgressPreference = 'SilentlyContinue' ; "
                    "$ErrorActionPreference = 'Stop' ; "
                    f"$response = Invoke-WebRequest -Uri '{url}' -UseBasicParsing ; "
                    f"[System.IO.File]::WriteAllBytes('{escaped_out_path}', $response.Content)"
                )  # fmt: skip
            else:
                command = (
                    "curl "
                    f"-o '{out_path}' "
                    f"'{url}'"
                    ";"
                    "touch "
                    f"'{out_path}'"
                )  # fmt: skip

        return command, put_args



#!/usr/bin/python
#
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


DOCUMENTATION = '''
---
module: azure_rm_cognitiveservicesmodel_info

version_added: "3.20.0"

short_description: List cognitiveservices models facts

description:
    - List all model details under location.

options:
    location:
        description:
            - The location of the models.
        type: str
        required: true

extends_documentation_fragment:
    - azure.azcollection.azure

author:
    - bpeck@redhat.com (@p3ck)
'''

EXAMPLES = '''
- name: List all models at location
  azure_rm_cognitiveservicesmodel_info:
    location: eastus
'''

RETURN = '''
models:
    description:
        - List of model details.
    returned: always
    type: dict
    sample: [
        {
            "kind": "OpenAI",
            "model": {
                "capabilities": {
                    "imageGenerations": "true",
                    "inference": "true"
                },
                "deprecation": {
                    "inference": "2026-03-04T00:00:00Z"
                },
                "format": "OpenAI",
                "is_default_version": true,
                "lifecycle_status": "Deprecated",
                "max_capacity": 2,
                "name": "dall-e-3",
                "skus": [
                    {
                        "capacity": {
                            "default": 1,
                            "maximum": 1000
                        },
                        "deprecation_date": "2026-03-04T00:00:00.000Z",
                        "name": "Standard",
                        "rate_limits": [
                            {
                                "count": 3.0,
                                "renewal_period": 60.0
                            }
                        ],
                        "usage_name": "OpenAI.Standard.Dalle"
                    }
                ],
                "system_data": {
                    "created_at": "2023-08-11T00:00:00.000Z",
                    "created_by": "Microsoft",
                    "last_modified_at": "2023-08-11T00:00:00.000Z",
                    "last_modified_by": "Microsoft"
                },
                "version": "3.0"
            },
            "sku_name": "S0"
        },
        {
            "kind": "OpenAI",
            "model": {
                "capabilities": {
                    "imageGenerations": "true",
                    "inference": "true"
                },
                "deprecation": {
                    "inference": "2025-01-27T00:00:00Z"
                },
                "format": "OpenAI",
                "is_default_version": true,
                "lifecycle_status": "Deprecated",
                "max_capacity": 2,
                "name": "dall-e-2",
                "skus": [
                    {
                        "capacity": {
                            "default": 1,
                            "maximum": 1000
                        },
                        "deprecation_date": "2025-01-27T00:00:00.000Z",
                        "name": "Standard",
                        "rate_limits": [
                            {
                                "count": 3.0,
                                "renewal_period": 60.0
                            }
                        ],
                        "usage_name": "OpenAI.Standard.Dalle-2"
                    }
                ],
                "system_data": {
                    "created_at": "2024-04-15T00:00:00.000Z",
                    "created_by": "Microsoft",
                    "last_modified_at": "2024-04-15T00:00:00.000Z",
                    "last_modified_by": "Microsoft"
                },
                "version": "2.0"
            },
            "sku_name": "S0"
        },
    ]
'''

from ansible_collections.azure.azcollection.plugins.module_utils.azure_rm_common import AzureRMModuleBase


class AzureRMCognitiveServicesModelInfo(AzureRMModuleBase):

    def __init__(self):

        self.module_arg_spec = dict(
            location=dict(
                type='str',
                required=True,
            ),
        )

        self.results = dict(
            changed=False,
            models=[],
        )

        super(AzureRMCognitiveServicesModelInfo, self).__init__(
            self.module_arg_spec,
            supports_check_mode=True,
            supports_tags=False,
            facts_module=True
        )

    def exec_module(self, **kwargs):
        for key in self.module_arg_spec:
            setattr(self, key, kwargs[key])

        self.results['models'] = self.list_all(self.location)

        return self.results

    def list_all(self, location):
        results = []
        try:
            models = self.cognitive_services_management_client.models.list(
                location=location
            )
            for model in models:
                results.append(model.as_dict())
        except Exception as exc:
            self.fail('Error when listing all models under subscription got Exception as {0}'.format(str(exc)))
        return results


def main():
    AzureRMCognitiveServicesModelInfo()


if __name__ == '__main__':
    main()



"""
textfsm parser

This is the textfsm parser for use with the cli_parse module and action plugin
https://github.com/google/textfsm
"""

from __future__ import absolute_import, division, print_function


__metaclass__ = type

DOCUMENTATION = """
    author: Bradley Thornton (@cidrblock)
    name: textfsm
    short_description: Define configurable options for C(textfsm) sub-plugin of M(ansible.utils.cli_parse) module
    description:
    - This plugin documentation provides the configurable options that can be passed
      to the M(ansible.utils.cli_parse) plugins when I(ansible.utils.textfsm) is used as a value for
      I(name) option.
    version_added: 1.0.0
"""

EXAMPLES = r"""
- name: "Run command and parse with textfsm"
  ansible.utils.cli_parse:
    command: "show version"
    parser:
      name: ansible.utils.textfsm
  register: nxos_textfsm_command

- name: "Pass text and command"
  ansible.utils.cli_parse:
    text: "{{ lookup('ansible.builtin.file', '/home/user/files/nxos_show_version.txt') }}"
    parser:
      name: ansible.utils.textfsm
      template_path: "/home/user/templates/nxos_show_version.textfsm"
  register: nxos_textfsm_text
"""

import os

from ansible.module_utils.basic import missing_required_lib
from ansible.module_utils.common.text.converters import to_native

from ansible_collections.ansible.utils.plugins.plugin_utils.base.cli_parser import CliParserBase


try:
    import textfsm

    HAS_TEXTFSM = True
except ImportError:
    HAS_TEXTFSM = False


class CliParser(CliParserBase):
    """The textfsm parser class
    Convert raw text to structured data using textfsm
    """

    DEFAULT_TEMPLATE_EXTENSION = "textfsm"
    PROVIDE_TEMPLATE_CONTENTS = False

    @staticmethod
    def _check_reqs():
        """Check the prerequisites for the textfsm parser

        :return dict: A dict with errors or a template_path
        """
        errors = []

        if not HAS_TEXTFSM:
            errors.append(missing_required_lib("textfsm"))

        return {"errors": errors}

    def parse(self, *_args, **_kwargs):
        """Std entry point for a cli_parse parse execution

        :return: Errors or parsed text as structured data
        :rtype: dict

        :example:

        The parse function of a parser should return a dict:
        {"errors": [a list of errors]}
        or
        {"parsed": obj}
        """
        cli_output = self._task_args.get("text")
        res = self._check_reqs()
        if res.get("errors"):
            return {"errors": res.get("errors")}

        template_path = self._task_args.get("parser").get("template_path")
        if template_path and not os.path.isfile(template_path):
            return {
                "errors": "error while reading template_path file {file}".format(
                    file=template_path,
                ),
            }
        try:
            template = open(self._task_args.get("parser").get("template_path"))
        except IOError as exc:
            return {"errors": to_native(exc)}

        re_table = textfsm.TextFSM(template)
        fsm_results = re_table.ParseText(cli_output)

        results = list()
        for item in fsm_results:
            results.append(dict(zip(re_table.header, item)))

        return {"parsed": results}



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or
# https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
module: networks_firmware_upgrades_staged_events
short_description: Resource module for networks _firmware _upgrades _staged _events
description:
  - Manage operations create and update of the resource networks _firmware _upgrades
    _staged _events.
  - Create a Staged Upgrade Event for a network.
  - Update the Staged Upgrade Event for a network.
version_added: '1.0.0'
extends_documentation_fragment:
  - cisco.meraki.module
author: Francisco Munoz (@fmunoz)
options:
  networkId:
    description: NetworkId path parameter. Network ID.
    type: str
  products:
    description: Contains firmware upgrade version information.
    suboptions:
      switch:
        description: Version information for the switch network being upgraded.
        suboptions:
          nextUpgrade:
            description: The next upgrade version for the switch network.
            suboptions:
              toVersion:
                description: The version to be updated to for switch devices.
                suboptions:
                  id:
                    description: The version ID.
                    type: str
                type: dict
            type: dict
        type: dict
      switchCatalyst:
        description: Version information for the switch network being upgraded.
        suboptions:
          nextUpgrade:
            description: The next upgrade version for the switch network.
            suboptions:
              toVersion:
                description: The version to be updated to for switch Catalyst devices.
                suboptions:
                  id:
                    description: The version ID.
                    type: str
                type: dict
            type: dict
        type: dict
    type: dict
  stages:
    description: All firmware upgrade stages in the network with their start time.
    elements: dict
    suboptions:
      group:
        description: The Staged Upgrade Group containing the name and ID.
        suboptions:
          id:
            description: ID of the Staged Upgrade Group.
            type: str
        type: dict
      milestones:
        description: The Staged Upgrade Milestones for the specific stage.
        suboptions:
          scheduledFor:
            description: The start time of the staged upgrade stage. (In ISO-8601
              format, in the time zone of the network.).
            type: str
        type: dict
    type: list
requirements:
  - meraki >= 2.4.9
  - python >= 3.5
seealso:
  - name: Cisco Meraki documentation for networks createNetworkFirmwareUpgradesStagedEvent
    description: Complete reference of the createNetworkFirmwareUpgradesStagedEvent
      API.
    link: https://developer.cisco.com/meraki/api-v1/#!create-network-firmware-upgrades-staged-event
  - name: Cisco Meraki documentation for networks updateNetworkFirmwareUpgradesStagedEvents
    description: Complete reference of the updateNetworkFirmwareUpgradesStagedEvents
      API.
    link: https://developer.cisco.com/meraki/api-v1/#!update-network-firmware-upgrades-staged-events
notes:
  - SDK Method used are
    networks.Networks.create_network_firmware_upgrades_staged_event,
    networks.Networks.update_network_firmware_upgrades_staged_events,
  - Paths used are
    post /networks/{networkId}/firmwareUpgrades/staged/events,
    put
    /networks/{networkId}/firmwareUpgrades/staged/events,
"""

EXAMPLES = r"""
- name: Create
  cisco.meraki.networks_firmware_upgrades_staged_events:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    state: present
    networkId: string
    products:
      switch:
        nextUpgrade:
          toVersion:
            id: '1234'
      switchCatalyst:
        nextUpgrade:
          toVersion:
            id: '4321'
    stages:
      - group:
          id: '1234'
        milestones:
          scheduledFor: '2018-02-11T00:00:00Z'
- name: Update all
  cisco.meraki.networks_firmware_upgrades_staged_events:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    state: present
    networkId: string
    stages:
      - group:
          id: '1234'
        milestones:
          scheduledFor: '2018-02-11T00:00:00Z'
"""
RETURN = r"""
meraki_response:
  description: A dictionary or list with the response returned by the Cisco Meraki Python SDK
  returned: always
  type: dict
  sample: >
    {
      "products": {
        "switch": {
          "nextUpgrade": {
            "toVersion": {
              "id": "string",
              "shortName": "string"
            }
          }
        }
      },
      "stages": [
        {
          "group": {
            "id": "string",
            "name": "string",
            "description": "string"
          },
          "milestones": {
            "scheduledFor": "string",
            "startedAt": "string",
            "completedAt": "string",
            "canceledAt": "string"
          },
          "status": "string"
        }
      ],
      "reasons": [
        {
          "category": "string",
          "comment": "string"
        }
      ]
    }
"""



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2018, Dag Wieers (@dagwieers) <dag@wieers.com>
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type

ANSIBLE_METADATA = {"metadata_version": "1.1", "status": ["deprecated"], "supported_by": "community"}

DOCUMENTATION = r"""
---
module: mso_schema_template_deploy
short_description: Deploy schema templates to sites
description:
- Deploy schema templates to sites.
- Prior to deploy a schema validation is executed for MSO releases running on the ND platform.
- When schema validation fails, M(cisco.mso.mso_schema_template_deploy) fails and deploy will not be executed.
author:
- Dag Wieers (@dagwieers)
options:
  schema:
    description:
    - The name of the schema.
    type: str
    required: true
  template:
    description:
    - The name of the template.
    type: str
    required: true
    aliases: [ name ]
  site:
    description:
    - The name of the site B(to undeploy).
    type: str
  state:
    description:
    - Use C(deploy) to deploy schema template.
    - Use C(status) to get deployment status.
    - Use C(undeploy) to deploy schema template from a site.
    type: str
    choices: [ deploy, status, undeploy ]
    default: deploy

deprecated:
  removed_in: '3.0.0'
  why: Due to changes in the ND and NDO API, a new module (cisco.mso.ndo_template_deploy) has been released for
    ND v2.2 (NDO v4.1) and later.
  alternative: Use M(cisco.mso.ndo_template_deploy) instead.
seealso:
- module: cisco.mso.mso_schema_site
- module: cisco.mso.mso_schema_template
extends_documentation_fragment: cisco.mso.modules
"""

EXAMPLES = r"""
- name: Deploy a schema template
  cisco.mso.mso_schema_template_deploy:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema 1
    template: Template 1
    state: deploy

- name: Undeploy a schema template
  cisco.mso.mso_schema_template_deploy:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema 1
    template: Template 1
    site: Site 1
    state: undeploy

- name: Get deployment status
  cisco.mso.mso_schema_template_deploy:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema 1
    template: Template 1
    state: status
  register: status_result
"""

RETURN = r"""
"""

from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.mso.plugins.module_utils.mso import MSOModule, mso_argument_spec


def main():
    argument_spec = mso_argument_spec()
    argument_spec.update(
        schema=dict(type="str", required=True),
        template=dict(type="str", required=True, aliases=["name"]),
        site=dict(type="str"),
        state=dict(type="str", default="deploy", choices=["deploy", "status", "undeploy"]),
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_if=[
            ["state", "undeploy", ["site"]],
        ],
    )

    schema = module.params.get("schema")
    template = module.params.get("template").replace(" ", "")
    site = module.params.get("site")
    state = module.params.get("state")

    mso = MSOModule(module)

    # Get schema id
    schema_id = mso.lookup_schema(schema)

    payload = dict(
        schemaId=schema_id,
        templateName=template,
    )

    qs = None
    if state == "deploy":
        if mso.platform == "nd":
            mso.validate_schema(schema_id)
        path = "execute/schema/{0}/template/{1}".format(schema_id, template)
    elif state == "status":
        path = "status/schema/{0}/template/{1}".format(schema_id, template)
    elif state == "undeploy":
        path = "execute/schema/{0}/template/{1}".format(schema_id, template)
        site_id = mso.lookup_site(site)
        qs = dict(undeploy=site_id)

    if not module.check_mode:
        status = mso.request(path, method="GET", data=payload, qs=qs)
        mso.exit_json(**status)
    else:
        mso.exit_json()


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# Copyright 2019 Red Hat Inc.
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
"""
The iosxr_l3_interfaces class
It is in this file where the current configuration (as dict)
is compared to the provided configuration (as dict) and the command set
necessary to bring the current configuration to it's desired end-state is
created
"""

from __future__ import absolute_import, division, print_function

import copy


__metaclass__ = type


from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.cfg.base import (
    ConfigBase,
)
from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.utils import to_list

from ansible_collections.cisco.iosxr.plugins.module_utils.network.iosxr.facts.facts import Facts
from ansible_collections.cisco.iosxr.plugins.module_utils.network.iosxr.utils.utils import (
    add_command_to_config_list,
    dict_to_set,
    filter_dict_having_none_value,
    normalize_interface,
    remove_command_from_config_list,
    remove_duplicate_interface,
    validate_ipv6,
    validate_n_expand_ipv4,
)


class L3_Interfaces(ConfigBase):
    """
    The iosxr_l3_interfaces class
    """

    gather_subset = ["!all", "!min"]

    gather_network_resources = ["l3_interfaces"]

    def get_l3_interfaces_facts(self, data=None):
        """Get the 'facts' (the current configuration)
        :rtype: A dictionary
        :returns: The current configuration as a dictionary
        """
        facts, _warnings = Facts(self._module).get_facts(
            self.gather_subset,
            self.gather_network_resources,
            data=data,
        )
        l3_interfaces_facts = facts["ansible_network_resources"].get(
            "l3_interfaces",
        )
        if not l3_interfaces_facts:
            return []
        return l3_interfaces_facts

    def execute_module(self):
        """Execute the module
        :rtype: A dictionary
        :returns: The result from module execution
        """
        result = {"changed": False}
        warnings = list()
        commands = list()

        if self.state in self.ACTION_STATES:
            existing_l3_interfaces_facts = self.get_l3_interfaces_facts()
        else:
            existing_l3_interfaces_facts = []

        if self.state in self.ACTION_STATES or self.state == "rendered":
            commands.extend(self.set_config(existing_l3_interfaces_facts))

        if commands and self.state in self.ACTION_STATES:
            if not self._module.check_mode:
                self._connection.edit_config(commands)
            result["changed"] = True

        if self.state in self.ACTION_STATES:
            result["commands"] = commands

        if self.state in self.ACTION_STATES or self.state == "gathered":
            changed_l3_interfaces_facts = self.get_l3_interfaces_facts()

        elif self.state == "rendered":
            result["rendered"] = commands

        elif self.state == "parsed":
            running_config = self._module.params["running_config"]
            if not running_config:
                self._module.fail_json(
                    msg="value of running_config parameter must not be empty for state parsed",
                )
            result["parsed"] = self.get_l3_interfaces_facts(
                data=running_config,
            )

        if self.state in self.ACTION_STATES:
            result["before"] = existing_l3_interfaces_facts
            if result["changed"]:
                result["after"] = changed_l3_interfaces_facts

        elif self.state == "gathered":
            result["gathered"] = changed_l3_interfaces_facts

        result["warnings"] = warnings
        return result

    def set_config(self, existing_l3_interfaces_facts):
        """Collect the configuration from the args passed to the module,
            collect the current configuration (as a dict from facts)
        :rtype: A list
        :returns: the commands necessary to migrate the current configuration
                  to the desired configuration
        """
        want = self._module.params["config"]
        have = copy.deepcopy(existing_l3_interfaces_facts)
        resp = self.set_state(want, have)
        return to_list(resp)

    def set_state(self, want, have):
        """Select the appropriate function based on the state provided
        :param want: the desired configuration as a dictionary
        :param have: the current configuration as a dictionary
        :rtype: A list
        :returns: the commands necessary to migrate the current configuration
                  to the desired configuration
        """
        commands = []

        if self.state in ("overridden", "merged", "replaced", "rendered") and not want:
            self._module.fail_json(
                msg="value of config parameter must not be empty for state {0}".format(
                    self.state,
                ),
            )

        if self.state == "overridden":
            commands = self._state_overridden(want, have, self._module)
        elif self.state == "deleted":
            commands = self._state_deleted(want, have)
        elif self.state in ("merged", "rendered"):
            commands = self._state_merged(want, have, self._module)
        elif self.state == "replaced":
            commands = self._state_replaced(want, have, self._module)

        return commands

    def _state_replaced(self, want, have, module):
        """The command generator when state is replaced
        :rtype: A list
        :returns: the commands necessary to migrate the current configuration
                  to the desired configuration
        """
        commands = []

        for interface in want:
            interface["name"] = normalize_interface(interface["name"])
            for each in have:
                if each["name"] == interface["name"]:
                    break
            else:
                commands.extend(self._set_config(interface, dict(), module))
                continue
            have_dict = filter_dict_having_none_value(interface, each)
            commands.extend(self._clear_config(dict(), have_dict))
            commands.extend(self._set_config(interface, each, module))
        # Remove the duplicate interface call
        commands = remove_duplicate_interface(commands)

        return commands

    def _state_overridden(self, want, have, module):
        """The command generator when state is overridden
        :rtype: A list
        :returns: the commands necessary to migrate the current configuration
                  to the desired configuration
        """
        commands = []
        not_in_have = set()
        in_have = set()

        for each in have:
            for interface in want:
                interface["name"] = normalize_interface(interface["name"])
                if each["name"] == interface["name"]:
                    in_have.add(interface["name"])
                    break
                if interface["name"] != each["name"]:
                    not_in_have.add(interface["name"])
            else:
                # We didn't find a matching desired state, which means we can
                # pretend we received an empty desired state.
                interface = dict(name=each["name"])
                kwargs = {"want": interface, "have": each}
                commands.extend(self._clear_config(**kwargs))
                continue
            have_dict = filter_dict_having_none_value(interface, each)
            commands.extend(self._clear_config(dict(), have_dict))
            commands.extend(self._set_config(interface, each, module))
        # Add the want interface that's not already configured in have interface
        for each in not_in_have - in_have:
            for every in want:
                interface = "interface {0}".format(every["name"])
                if each and interface not in commands:
                    commands.extend(self._set_config(every, {}, module))
        # Remove the duplicate interface call
        commands = remove_duplicate_interface(commands)

        return commands

    def _state_merged(self, want, have, module):
        """The command generator when state is merged
        :rtype: A list
        :returns: the commands necessary to merge the provided into
                  the current configuration
        """
        commands = []

        for interface in want:
            interface["name"] = normalize_interface(interface["name"])
            if self.state == "rendered":
                commands.extend(self._set_config(interface, dict(), module))
            else:
                for each in have:
                    if each["name"] == interface["name"]:
                        break
                else:
                    commands.extend(
                        self._set_config(interface, dict(), module),
                    )
                    continue
                commands.extend(self._set_config(interface, each, module))

        return commands

    def _state_deleted(self, want, have):
        """The command generator when state is deleted
        :rtype: A list
        :returns: the commands necessary to remove the current configuration
                  of the provided objects
        """
        commands = []

        if want:
            for interface in want:
                interface["name"] = normalize_interface(interface["name"])
                for each in have:
                    if each["name"] == interface["name"] or interface["name"] in each["name"]:
                        break
                else:
                    continue
                interface = dict(name=interface["name"])
                commands.extend(self._clear_config(interface, each))
        else:
            for each in have:
                want = dict()
                commands.extend(self._clear_config(want, each))

        return commands

    def verify_diff_again(self, want, have):
        """
        Verify the IPV4 difference again as sometimes due to
        change in order of set, set difference may result into change,
        when there's actually no difference between want and have
        :param want: want_dict IPV4
        :param have: have_dict IPV4
        :return: diff
        """
        diff = False
        for each in want:
            each_want = dict(each)
            for every in have:
                every_have = dict(every)

                if each_want.get("address") == every_have.get("address"):
                    if len(each_want.keys()) == len(every_have.keys()) and (
                        each_want.get("secondary") == every_have.get("secondary")
                    ):
                        diff = False
                        break
                    if not each_want.get("secondary") and not every_have.get(
                        "secondary",
                    ):
                        diff = False
                        break

                    diff = True
                else:
                    diff = True
            if diff:
                break

        return diff

    def _expand_flow_directions(self, direction):
        if direction == "bidirectional":
            return ["ingress", "egress"]
        return [direction]

    def _add_no_flow_command(self, interface, protocol, flow_cfg, commands):
        for direction in self._expand_flow_directions(flow_cfg["direction"]):
            cmd = "no flow {0} monitor {1} sampler {2} {3}".format(
                protocol,
                flow_cfg["monitor"],
                flow_cfg["sampler"],
                direction,
            )
            add_command_to_config_list(interface, cmd, commands)

    def _add_flow_command(self, interface, protocol, flow_cfg, commands):
        for direction in self._expand_flow_directions(flow_cfg["direction"]):
            cmd = "flow {0} monitor {1} sampler {2} {3}".format(
                protocol,
                flow_cfg["monitor"],
                flow_cfg["sampler"],
                direction,
            )
            add_command_to_config_list(interface, cmd, commands)

    def _set_config(self, want, have, module):
        # Set the interface config based on the want and have config
        commands = []
        interface = "interface " + want["name"]

        # To handle L3 IPV4 configuration
        if want.get("ipv4"):
            for each in want.get("ipv4"):
                if each.get("address") != "dhcp":
                    each["address"] = validate_n_expand_ipv4(module, each)

        if have.get("ipv4"):
            for each in have.get("ipv4"):
                if each.get("address") != "dhcp":
                    each["address"] = validate_n_expand_ipv4(module, each)

        # Temporarily remove 'flow' before comparison
        want_flow = want.pop("flow", None)
        have_flow = have.pop("flow", None)

        # Get the diff b/w want and have
        want_dict = dict_to_set(want)
        have_dict = dict_to_set(have)

        # To handle L3 IPV4 configuration
        want_ipv4 = dict(want_dict).get("ipv4")
        have_ipv4 = dict(have_dict).get("ipv4")
        if want_ipv4:
            if have_ipv4:
                diff_ipv4 = set(want_ipv4) - set(dict(have_dict).get("ipv4"))
                if diff_ipv4:
                    diff_ipv4 = diff_ipv4 if self.verify_diff_again(want_ipv4, have_ipv4) else ()
            else:
                diff_ipv4 = set(want_ipv4)
            for each in diff_ipv4:
                ipv4_dict = dict(each)
                if ipv4_dict.get("address") != "dhcp":
                    cmd = "ipv4 address {0}".format(ipv4_dict["address"])
                    if ipv4_dict.get("secondary"):
                        cmd += " secondary"
                    add_command_to_config_list(interface, cmd, commands)

        # To handle L3 IPV6 configuration
        want_ipv6 = dict(want_dict).get("ipv6")
        have_ipv6 = dict(have_dict).get("ipv6")
        if want_ipv6:
            if have_ipv6:
                diff_ipv6 = set(want_ipv6) - set(have_ipv6)
            else:
                diff_ipv6 = set(want_ipv6)
            for each in diff_ipv6:
                ipv6_dict = dict(each)
                validate_ipv6(ipv6_dict.get("address"), module)
                cmd = "ipv6 address {0}".format(ipv6_dict.get("address"))
                add_command_to_config_list(interface, cmd, commands)

        if want.get("load_interval"):
            if want["load_interval"] != have.get("load_interval"):
                cmd = "load-interval {0}".format(want["load_interval"])
                add_command_to_config_list(interface, cmd, commands)

        if want.get("flow_control"):
            if want["flow_control"] != have.get("flow_control"):
                cmd = "flow-control {0}".format(want["flow_control"])
                add_command_to_config_list(interface, cmd, commands)

        want_cd = want.get("carrier_delay")
        have_cd = have.get("carrier_delay")

        if want_cd and want_cd != have_cd:
            cmd_parts = ["carrier-delay"]
            want_up = want_cd.get("up")
            want_down = want_cd.get("down")
            have_up = have_cd.get("up") if have_cd else None
            have_down = have_cd.get("down") if have_cd else None

            if want_up is not None and want_up != have_up:
                cmd_parts.append("up {}".format(want_up))
            if want_down is not None and want_down != have_down:
                cmd_parts.append("down {}".format(want_down))

            if len(cmd_parts) > 1:
                cmd = " ".join(cmd_parts)
                add_command_to_config_list(interface, cmd, commands)

        dampening_want = want.get("dampening")
        dampening_have = have.get("dampening")

        if dampening_want and dampening_want != dampening_have:
            if dampening_want.get("enabled"):
                if dampening_want.get("half_life") is None:
                    cmd = "dampening"
                else:
                    cmd_parts = ["dampening"]
                    params_order = [
                        "half_life",
                        "reuse_threshold",
                        "suppress_threshold",
                        "max_suppress_time",
                        "restart_penalty",
                    ]
                    for param in params_order:
                        value = dampening_want.get(param)
                        if value is not None:
                            cmd_parts.append(str(value))
                        else:
                            break
                    cmd = " ".join(cmd_parts)
                add_command_to_config_list(interface, cmd, commands)

        if want_flow is not None or have_flow is not None:
            have_flow = have_flow or {}
            want_flow = want_flow or {}

            if self.state == "replaced" and have_flow:
                for proto in ["ipv4", "ipv6"]:
                    if proto in have_flow and proto not in want_flow:
                        have_cfg = have_flow[proto]
                        self._add_no_flow_command(interface, proto, have_cfg, commands)

            for proto, want_cfg in want_flow.items():
                if want_cfg is None:
                    continue

                have_cfg = have_flow.get(proto, {})
                if want_cfg != have_cfg:
                    if have_cfg:
                        self._add_no_flow_command(interface, proto, have_cfg, commands)
                    self._add_flow_command(interface, proto, want_cfg, commands)

        if want_flow is not None:
            want["flow"] = want_flow
        if have_flow is not None:
            have["flow"] = have_flow

        return commands

    def _clear_config(self, want, have):
        # Delete the interface config based on the want and have config
        count = 0
        commands = []
        if want.get("name"):
            interface = "interface " + want["name"]
        else:
            interface = "interface " + have["name"]

        if have.get("ipv4") and want.get("ipv4"):
            for each in have.get("ipv4"):
                if each.get("secondary") and not (want.get("ipv4")[count].get("secondary")):
                    cmd = "ipv4 address {0} secondary".format(
                        each.get("address"),
                    )
                    remove_command_from_config_list(interface, cmd, commands)
                count += 1
        if have.get("ipv4") and not (want.get("ipv4")):
            remove_command_from_config_list(
                interface,
                "ipv4 address",
                commands,
            )
        if have.get("ipv6") and not (want.get("ipv6")):
            remove_command_from_config_list(
                interface,
                "ipv6 address",
                commands,
            )

        if have.get("carrier_delay") and not (want.get("carrier_delay")):
            remove_command_from_config_list(interface, "carrier-delay", commands)

        if have.get("dampening") and not (want.get("dampening")):
            remove_command_from_config_list(interface, "dampening", commands)

        if have.get("load_interval") and not want.get("load_interval"):
            remove_command_from_config_list(interface, "load-interval", commands)

        if have.get("flow_control") and not want.get("flow_control"):
            remove_command_from_config_list(interface, "flow-control", commands)

        if have.get("flow") and not want.get("flow"):
            for proto in ["ipv4", "ipv6"]:
                if have["flow"].get(proto):
                    self._add_no_flow_command(interface, proto, have["flow"][proto], commands)

        return commands



#
# -*- coding: utf-8 -*-
# Copyright 2019 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
"""
The iosxr_l3_interfaces fact class
It is in this file the configuration is collected from the device
for a given resource, parsed, and the facts tree is populated
based on the configuration.
"""

from __future__ import absolute_import, division, print_function


__metaclass__ = type


import re

from copy import deepcopy

from ansible_collections.ansible.netcommon.plugins.module_utils.network.common import utils

from ansible_collections.cisco.iosxr.plugins.module_utils.network.iosxr.argspec.l3_interfaces.l3_interfaces import (
    L3_InterfacesArgs,
)
from ansible_collections.cisco.iosxr.plugins.module_utils.network.iosxr.utils.utils import (
    get_interface_type,
    netmask_to_cidr,
)


class L3_InterfacesFacts(object):
    """The iosxr_l3_interfaces fact class"""

    def __init__(self, module, subspec="config", options="options"):
        self._module = module
        self.argument_spec = L3_InterfacesArgs.argument_spec
        spec = deepcopy(self.argument_spec)
        if subspec:
            if options:
                facts_argument_spec = spec[subspec][options]
            else:
                facts_argument_spec = spec[subspec]
        else:
            facts_argument_spec = spec

        self.generated_spec = utils.generate_dict(facts_argument_spec)

    def get_config(self, connection):
        return connection.get_config(flags="interface")

    def populate_facts(self, connection, ansible_facts, data=None):
        """Populate the facts for interfaces
        :param connection: the device connection
        :param ansible_facts: Facts dictionary
        :param data: previously collected conf
        :rtype: dictionary
        :returns: facts
        """
        objs = []

        if not data:
            data = self.get_config(connection)
        # operate on a collection of resource x
        config = ("\n" + data).split("\ninterface ")
        for conf in config:
            if conf:
                obj = self.render_config(self.generated_spec, conf)
                if obj:
                    objs.append(obj)
        facts = {}

        if objs:
            facts["l3_interfaces"] = []
            params = utils.validate_config(
                self.argument_spec,
                {"config": objs},
            )
            for cfg in params["config"]:
                facts["l3_interfaces"].append(utils.remove_empties(cfg))
        ansible_facts["ansible_network_resources"].update(facts)

        return ansible_facts

    def render_config(self, spec, conf):
        """
        Render config as dictionary structure and delete keys from spec for null values
        :param spec: The facts tree, generated from the argspec
        :param conf: The configuration
        :rtype: dictionary
        :returns: The generated config
        """
        config = deepcopy(spec)
        match = re.search(r"^(\S+)", conf)
        if match:
            intf = match.group(1)
            if match.group(1).lower() == "preconfigure":
                match = re.search(r"^(\S+) (.*)", conf)
                if match:
                    intf = match.group(2)

            if get_interface_type(intf) == "unknown":
                return {}

            # populate the facts from the configuration
            config["name"] = intf

            # Get the configured IPV4 details
            ipv4 = []
            ipv4_all = re.findall(r"ipv4 address (\S+.*)", conf)
            for each in ipv4_all:
                each_ipv4 = dict()
                if "secondary" in each:
                    each_ipv4["address"] = self.format_ipv4(each.split(" secondary")[0])
                    each_ipv4["secondary"] = True
                elif "secondary" not in each and "dhcp" not in each:
                    each_ipv4["address"] = self.format_ipv4(each)
                elif "dhcp" in each:
                    each_ipv4["address"] = "dhcp"
                ipv4.append(each_ipv4)
                config["ipv4"] = ipv4

            # Get the configured IPV6 details
            ipv6 = []
            ipv6_all = re.findall(r"ipv6 address (\S+)", conf)
            for each in ipv6_all:
                each_ipv6 = dict()
                each_ipv6["address"] = each
                ipv6.append(each_ipv6)
                config["ipv6"] = ipv6

            carrier_delay_match = re.search(r"^\s*carrier-delay (.*)$", conf, re.M)
            if carrier_delay_match:
                args = carrier_delay_match.group(1).strip()
                up_match = re.search(r"up (\d+)", args)
                if up_match:
                    config["carrier_delay"]["up"] = int(up_match.group(1))

                down_match = re.search(r"down (\d+)", args)
                if down_match:
                    config["carrier_delay"]["down"] = int(down_match.group(1))

            dampening_line = re.search(r"^\s*dampening(.*)$", conf, re.M)
            if dampening_line:
                config["dampening"]["enabled"] = True

                params = dampening_line.group(1).strip().split()

                param_keys = [
                    "half_life",
                    "reuse_threshold",
                    "suppress_threshold",
                    "max_suppress_time",
                    "restart_penalty",
                ]
                for i, value in enumerate(params):
                    if i < len(param_keys):
                        key_name = param_keys[i]
                        config["dampening"][key_name] = int(value)

            load_interval = re.search(r"load-interval (\d+)", conf)
            if load_interval:
                config["load_interval"] = int(load_interval.group(1))

            flow_control = re.search(r"flow(?:-control)? (ingress|egress|bidirectional)", conf)
            if flow_control:
                config["flow_control"] = flow_control.group(1)

            flow_monitor_lines = re.findall(
                r"^\s*flow (ipv4|ipv6) monitor (\S+) sampler (\S+) (ingress|egress)",
                conf,
                re.M,
            )
            if flow_monitor_lines:
                if "flow" not in config or not config["flow"]:
                    config["flow"] = {}

                flow_monitor_data = {}
                for line in flow_monitor_lines:
                    protocol, monitor, sampler, direction = line
                    if protocol not in flow_monitor_data:
                        flow_monitor_data[protocol] = {
                            "monitor": monitor,
                            "sampler": sampler,
                            "directions": set(),
                        }

                    proto_flow = flow_monitor_data[protocol]
                    if proto_flow["monitor"] != monitor or proto_flow["sampler"] != sampler:
                        proto_flow["monitor"] = monitor
                        proto_flow["sampler"] = sampler
                        proto_flow["directions"] = set()

                    proto_flow["directions"].add(direction)

                for protocol, proto_flow in flow_monitor_data.items():
                    directions = proto_flow["directions"]
                    if "ingress" in directions and "egress" in directions:
                        direction = "bidirectional"
                    elif "ingress" in directions:
                        direction = "ingress"
                    else:
                        direction = "egress"

                    config["flow"][protocol] = {
                        "monitor": proto_flow["monitor"],
                        "sampler": proto_flow["sampler"],
                        "direction": direction,
                    }

            return utils.remove_empties(config)

    def format_ipv4(self, address):
        parts = address.split(" ")
        if len(parts) > 1 and parts[1]:
            cidr_val = netmask_to_cidr(parts[1])
            return "{0}/{1}".format(parts[0], cidr_val)
        return parts[0]



#!/usr/bin/python
# -*- coding: utf-8 -*-

# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}

DOCUMENTATION = r'''
---
module: intersight_pool_reservation
short_description: Manage pool identifier reservations in Cisco Intersight
description:
  - Reserve and release identifiers (IP, UUID, MAC, WWNN, WWPN) from Intersight pools.
  - Handles the Intersight reservation API's one-time-use constraint by performing a pre-flight
    check before attempting to reserve, ensuring idempotent playbook runs.
  - Reserved identifiers are consumed when a policy using them is attached to a server profile
    or when the server profile is deployed.
  - Reservation of identifiers is only supported for Fabric Interconnect-attached servers.
  - For more information see L(Cisco Intersight,https://intersight.com/apidocs).
extends_documentation_fragment: intersight
options:
  pool_type:
    description:
      - The type of pool to manage reservations for.
    type: str
    required: true
    choices: [ip, uuid, mac, wwnn, wwpn]
  pool_name:
    description:
      - Name of the pool to reserve from.
    type: str
    required: true
  organization:
    description:
      - The name of the Organization the pool belongs to.
    type: str
    default: default
  identity:
    description:
      - The specific identifier value to reserve (e.g., an IP address, UUID, MAC address).
      - Required when C(state) is C(present).
    type: str
  state:
    description:
      - If C(present), will reserve the identifier in the pool.
      - If C(absent), will release the reservation.
    type: str
    choices: [present, absent]
    default: present
  allocation_type:
    description:
      - Whether the reservation is C(static) (specific value) or C(dynamic) (next available from pool).
      - When C(static), C(identity) is required.
      - When C(dynamic), the pool assigns the next available identifier.
    type: str
    choices: [static, dynamic]
    default: static
author:
  - Steve Fulmer (@stevefulme1)
'''

EXAMPLES = r'''
- name: Reserve a specific IP address from a pool
  cisco.intersight.intersight_pool_reservation:
    api_private_key: "{{ api_private_key }}"
    api_key_id: "{{ api_key_id }}"
    pool_type: ip
    pool_name: IP-Pool-01
    identity: "10.10.10.100"

- name: Reserve a specific UUID from a pool
  cisco.intersight.intersight_pool_reservation:
    api_private_key: "{{ api_private_key }}"
    api_key_id: "{{ api_key_id }}"
    pool_type: uuid
    pool_name: UUID-Pool-01
    identity: "550e8400-e29b-41d4-a716-446655440000"

- name: Reserve a MAC address
  cisco.intersight.intersight_pool_reservation:
    api_private_key: "{{ api_private_key }}"
    api_key_id: "{{ api_key_id }}"
    pool_type: mac
    pool_name: MAC-Pool-01
    identity: "00:25:B5:00:00:01"

- name: Release a reservation
  cisco.intersight.intersight_pool_reservation:
    api_private_key: "{{ api_private_key }}"
    api_key_id: "{{ api_key_id }}"
    pool_type: ip
    pool_name: IP-Pool-01
    identity: "10.10.10.100"
    state: absent

- name: Reserve next available IP (dynamic)
  cisco.intersight.intersight_pool_reservation:
    api_private_key: "{{ api_private_key }}"
    api_key_id: "{{ api_key_id }}"
    pool_type: ip
    pool_name: IP-Pool-01
    allocation_type: dynamic
'''

RETURN = r'''
api_response:
  description: The API response from the reservation operation.
  returned: always
  type: dict
  sample:
    "api_response": {
        "Identity": "10.10.10.100",
        "Moid": "63a1b2c3d4e5f6a7b8c9d0e1",
        "AllocationType": "static"
    }
'''


from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.intersight.plugins.module_utils.intersight import IntersightModule, intersight_argument_spec


POOL_TYPE_MAP = {
    'ip': {
        'resource_path': '/ippool/Reservations',
        'pool_path': '/ippool/Pools',
        'identity_key': 'Identity',
        'object_type': 'ippool.Reservation',
    },
    'uuid': {
        'resource_path': '/uuidpool/Reservations',
        'pool_path': '/uuidpool/Pools',
        'identity_key': 'Identity',
        'object_type': 'uuidpool.Reservation',
    },
    'mac': {
        'resource_path': '/macpool/Reservations',
        'pool_path': '/macpool/Pools',
        'identity_key': 'Identity',
        'object_type': 'macpool.Reservation',
    },
    'wwnn': {
        'resource_path': '/fcpool/Reservations',
        'pool_path': '/fcpool/Pools',
        'identity_key': 'Identity',
        'object_type': 'fcpool.Reservation',
    },
    'wwpn': {
        'resource_path': '/fcpool/Reservations',
        'pool_path': '/fcpool/Pools',
        'identity_key': 'Identity',
        'object_type': 'fcpool.Reservation',
    },
}


def get_pool_moid(intersight, pool_path, pool_name, organization_moid):
    """Look up a pool by name and org."""
    options = {
        'http_method': 'get',
        'resource_path': pool_path,
        'query_params': {
            '$filter': "Name eq '" + pool_name + "' and Organization.Moid eq '" + organization_moid + "'",
        },
    }
    response = intersight.call_api(**options)
    if response.get('Results'):
        return response['Results'][0]['Moid']
    return None


def get_existing_reservation(intersight, resource_path, pool_moid, identity, organization_moid):
    """Check if a reservation already exists for this identity or pool."""
    filter_str = "Organization.Moid eq '" + organization_moid + "'"
    if identity:
        filter_str += " and Identity eq '" + identity + "'"
    else:
        filter_str += " and Pool.Moid eq '" + pool_moid + "'"
    options = {
        'http_method': 'get',
        'resource_path': resource_path,
        'query_params': {
            '$filter': filter_str,
        },
    }
    response = intersight.call_api(**options)
    if response.get('Results'):
        return response['Results'][0]
    return None


def create_reservation(intersight, resource_path, pool_moid, organization_moid, identity, allocation_type, object_type, pool_type):
    """Create a new reservation."""
    body = {
        'ObjectType': object_type,
        'AllocationType': allocation_type,
        'Organization': {
            'Moid': organization_moid,
            'ObjectType': 'organization.Organization',
        },
    }
    if allocation_type == 'dynamic':
        body['Pool'] = {
            'Moid': pool_moid,
            'ObjectType': POOL_TYPE_MAP[pool_type]['object_type'].replace('Reservation', 'Pool'),
        }
    if identity and allocation_type == 'static':
        body['Identity'] = identity

    if pool_type == 'wwnn':
        body['IdPurpose'] = 'WWNN'
    elif pool_type == 'wwpn':
        body['IdPurpose'] = 'WWPN'

    options = {
        'http_method': 'post',
        'resource_path': resource_path,
        'body': body,
    }
    return intersight.call_api(**options)


def delete_reservation(intersight, resource_path, reservation_moid):
    """Delete a reservation."""
    options = {
        'http_method': 'delete',
        'resource_path': resource_path,
        'moid': reservation_moid,
    }
    return intersight.call_api(**options)


def main():
    argument_spec = intersight_argument_spec.copy()
    argument_spec.update(
        pool_type=dict(type='str', required=True, choices=['ip', 'uuid', 'mac', 'wwnn', 'wwpn']),
        pool_name=dict(type='str', required=True),
        organization=dict(type='str', default='default'),
        identity=dict(type='str'),
        state=dict(type='str', choices=['present', 'absent'], default='present'),
        allocation_type=dict(type='str', choices=['static', 'dynamic'], default='static'),
    )

    module = AnsibleModule(
        argument_spec,
        supports_check_mode=True,
        required_if=[
            ['allocation_type', 'static', ['identity']],
            ['state', 'absent', ['identity']],
        ],
    )

    intersight = IntersightModule(module)
    intersight.result['api_response'] = {}
    intersight.result['trace_id'] = ''

    pool_type = module.params['pool_type']
    pool_name = module.params['pool_name']
    organization_name = module.params['organization']
    identity = module.params['identity']
    state = module.params['state']
    allocation_type = module.params['allocation_type']

    pool_config = POOL_TYPE_MAP[pool_type]

    organization_moid = intersight.get_moid_by_name(
        resource_path='/organization/Organizations',
        resource_name=organization_name,
    )
    if not organization_moid:
        module.fail_json(msg="Organization '%s' not found." % organization_name)

    pool_moid = get_pool_moid(intersight, pool_config['pool_path'], pool_name, organization_moid)
    if not pool_moid:
        module.fail_json(msg="%s pool '%s' not found in organization '%s'." % (pool_type.upper(), pool_name, organization_name))

    existing = None
    if identity:
        existing = get_existing_reservation(intersight, pool_config['resource_path'], pool_moid, identity, organization_moid)

    if state == 'present':
        if existing:
            intersight.result['api_response'] = existing
            module.exit_json(**intersight.result)

        if module.check_mode:
            intersight.result['changed'] = True
            module.exit_json(**intersight.result)

        response = create_reservation(
            intersight,
            pool_config['resource_path'],
            pool_moid,
            organization_moid,
            identity,
            allocation_type,
            pool_config['object_type'],
            pool_type,
        )
        intersight.result['changed'] = True
        intersight.result['api_response'] = response

    elif state == 'absent':
        if not existing:
            module.exit_json(**intersight.result)

        if module.check_mode:
            intersight.result['changed'] = True
            module.exit_json(**intersight.result)

        delete_reservation(intersight, pool_config['resource_path'], existing['Moid'])
        intersight.result['changed'] = True

    module.exit_json(**intersight.result)


if __name__ == '__main__':
    main()



# Copyright (c) 2012, Michael DeHaan, <michael.dehaan@gmail.com>
# Copyright (c) 2017 Ansible Project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Make coding more python3-ish
from __future__ import annotations

DOCUMENTATION = r"""
author: Unknown (!UNKNOWN)
name: say
type: notification
requirements:
  - whitelisting in configuration
  - the C(/usr/bin/say) command line program (standard on macOS) or C(espeak) command line program
short_description: Notify using software speech synthesizer
description:
  - This plugin uses C(say) or C(espeak) to "speak" about play events.
"""

import os
import platform
import subprocess

from ansible.module_utils.common.process import get_bin_path
from ansible.plugins.callback import CallbackBase


class CallbackModule(CallbackBase):
    """
    makes Ansible much more exciting.
    """

    CALLBACK_VERSION = 2.0
    CALLBACK_TYPE = "notification"
    CALLBACK_NAME = "community.general.say"
    CALLBACK_NEEDS_WHITELIST = True

    def __init__(self):
        super().__init__()

        self.FAILED_VOICE = None
        self.REGULAR_VOICE = None
        self.HAPPY_VOICE = None
        self.LASER_VOICE = None

        try:
            self.synthesizer = get_bin_path("say")
            if platform.system() != "Darwin":
                # 'say' binary available, it might be GNUstep tool which doesn't support 'voice' parameter
                self._display.warning(
                    f"'say' executable found but system is '{platform.system()}': ignoring voice parameter"
                )
            else:
                self.FAILED_VOICE = "Zarvox"
                self.REGULAR_VOICE = "Trinoids"
                self.HAPPY_VOICE = "Cellos"
                self.LASER_VOICE = "Princess"
        except ValueError:
            try:
                self.synthesizer = get_bin_path("espeak")
                self.FAILED_VOICE = "klatt"
                self.HAPPY_VOICE = "f5"
                self.LASER_VOICE = "whisper"
            except ValueError:
                self.synthesizer = None

        # plugin disable itself if say is not present
        # ansible will not call any callback if disabled is set to True
        if not self.synthesizer:
            self.disabled = True
            self._display.warning(
                f"Unable to find either 'say' or 'espeak' executable, plugin {os.path.basename(__file__)} disabled"
            )

    def say(self, msg, voice):
        cmd = [self.synthesizer, msg]
        if voice:
            cmd.extend(("-v", voice))
        subprocess.call(cmd)

    def runner_on_failed(self, host, res, ignore_errors=False):
        self.say(f"Failure on host {host}", self.FAILED_VOICE)

    def runner_on_ok(self, host, res):
        self.say("pew", self.LASER_VOICE)

    def runner_on_skipped(self, host, item=None):
        self.say("pew", self.LASER_VOICE)

    def runner_on_unreachable(self, host, res):
        self.say(f"Failure on host {host}", self.FAILED_VOICE)

    def runner_on_async_ok(self, host, res, jid):
        self.say("pew", self.LASER_VOICE)

    def runner_on_async_failed(self, host, res, jid):
        self.say(f"Failure on host {host}", self.FAILED_VOICE)

    def playbook_on_start(self):
        self.say("Running Playbook", self.REGULAR_VOICE)

    def playbook_on_notify(self, host, handler):
        self.say("pew", self.LASER_VOICE)

    def playbook_on_task_start(self, name, is_conditional):
        if not is_conditional:
            self.say(f"Starting task: {name}", self.REGULAR_VOICE)
        else:
            self.say(f"Notifying task: {name}", self.REGULAR_VOICE)

    def playbook_on_setup(self):
        self.say("Gathering facts", self.REGULAR_VOICE)

    def playbook_on_play_start(self, name):
        self.say(f"Starting play: {name}", self.HAPPY_VOICE)

    def playbook_on_stats(self, stats):
        self.say("Play complete", self.HAPPY_VOICE)



# -*- coding: utf-8 -*-

# Copyright: Contributors to the Ansible project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

"""
Windows command execution via S3 for AWS SSM connections.

This module provides S3-based command execution to avoid PTY echo timeouts on Windows.
The PTY echoes ALL stdin data to stdout before PowerShell can read it, causing timeouts
on large commands. This approach keeps the session command tiny by uploading the actual
command to S3 and executing a small wrapper that downloads and runs it.

WARNING: ENCODING FLOW:
1. PowerShell -EncodedCommand contains UTF-16LE base64-encoded scripts
2. Decode -EncodedCommand to plain text (if present)
3. Upload plain text script to S3 with ContentType="text/plain; charset=utf-8"
4. PowerShell wrapper sets WebClient.Encoding to UTF8 before downloading
5. WriteAllText saves script file with UTF8 encoding
6. PowerShell executes the script file

All steps MUST use UTF-8 (except step 1 which decodes FROM UTF-16LE).
Breaking this chain causes Unicode corruption and module failures.
See: https://github.com/ansible-collections/amazon.aws/pull/2820
"""

from __future__ import annotations

import base64
import re
import typing
import uuid

if typing.TYPE_CHECKING:
    from typing import Any
    from typing import Callable
    from typing import Optional

from ansible.module_utils.common.text.converters import to_bytes
from ansible.module_utils.common.text.converters import to_text

from ansible_collections.amazon.aws.plugins.plugin_utils.ssm.common import S3_DOWNLOAD_ERROR_EXIT_CODE


class WindowsCommandExecutor:
    """Executes Windows PowerShell commands via S3 upload/download to avoid PTY echo issues."""

    def __init__(
        self,
        s3_client: Any,
        s3_manager: Any,
        session_manager: Any,
        exec_communicate: Callable,
        instance_id: str,
        bucket_name: str,
        verbosity_display: Callable[[int, str], None],
    ) -> None:
        """
        Initialise the WindowsCommandExecutor.

        :param s3_client: Boto3 S3 client for uploading/deleting objects
        :param s3_manager: S3ClientManager for generating presigned URLs
        :param session_manager: SSMSessionManager for writing to session stdin
        :param exec_communicate: Connection's exec_communicate method for executing commands
        :param instance_id: EC2 instance ID (used for S3 key prefixes)
        :param bucket_name: S3 bucket name for temporary command storage
        :param verbosity_display: Function for logging at different verbosity levels
        """
        self._s3_client = s3_client
        self._s3_manager = s3_manager
        self._session_manager = session_manager
        self._exec_communicate = exec_communicate
        self._instance_id = instance_id
        self._bucket_name = bucket_name
        self._verbosity_display = verbosity_display

    def execute(self, cmd: str, in_data: Optional[bytes], mark_begin: str, mark_end: str) -> tuple[int, str, str]:
        """
        Execute a Windows command by uploading to S3 and downloading via tiny wrapper command.

        This avoids PTY echo timeouts by keeping the session command tiny (~100 bytes)
        while the actual command (which may be kilobytes) is retrieved from S3.

        :param cmd: The PowerShell command to execute (may be encoded command line).
        :param in_data: Optional stdin data for the command.
        :param mark_begin: The begin marker for output parsing.
        :param mark_end: The end marker for output parsing.
        :returns: A tuple of (exit_code, stdout, stderr).
        """
        # Generate unique S3 key for this command
        command_id = str(uuid.uuid4())
        s3_key = f"{self._instance_id}/commands/{command_id}.ps1"
        stdin_key = None  # Initialize for cleanup scope

        self._verbosity_display(3, f"EXEC_VIA_S3: Uploading command to s3://{self._bucket_name}/{s3_key}")
        self._verbosity_display(4, f"EXEC_VIA_S3: Command length: {len(cmd)} bytes")

        # DEBUG: Show command encoding details (level 6 for long-term debugging of encoding issues)
        cmd_str = to_text(cmd)
        self._verbosity_display(6, f"EXEC_VIA_S3: Command (str): {repr(cmd_str[:200])}")
        if isinstance(cmd, bytes):
            self._verbosity_display(6, f"EXEC_VIA_S3: Command (bytes): {cmd[:200]}")
        self._verbosity_display(6, f"EXEC_VIA_S3: Command content:\n{cmd_str}")

        # Extract PowerShell script from -EncodedCommand if present
        script_to_upload = self._decode_powershell_command(cmd_str)

        try:
            # Upload script to S3
            presigned_url = self._upload_script_to_s3(s3_key, script_to_upload)

            # Handle stdin data if present
            if in_data:
                stdin_key = f"{self._instance_id}/commands/{command_id}-stdin.txt"
                stdin_url = self._upload_stdin_to_s3(stdin_key, in_data)
                wrapper = self._generate_wrapper_with_stdin(presigned_url, stdin_url, mark_begin, mark_end)
            else:
                wrapper = self._generate_wrapper_without_stdin(presigned_url, mark_begin, mark_end)

            # Execute wrapper command
            result = self._execute_wrapper(wrapper, mark_begin, mark_end)

            self._verbosity_display(3, f"EXEC_VIA_S3: Command execution complete, returncode={result[0]}")
            self._verbosity_display(6, f"EXEC_VIA_S3: Stdout (str): {repr(result[1][:200])}")
            self._verbosity_display(6, f"EXEC_VIA_S3: Stderr (str): {repr(result[2][:200])}")
            self._verbosity_display(6, f"EXEC_VIA_S3: Stdout type: {type(result[1])}, length: {len(result[1])}")
            self._verbosity_display(6, f"EXEC_VIA_S3: Stderr type: {type(result[2])}, length: {len(result[2])}")
            return result

        finally:
            # Clean up S3 objects
            self._cleanup_s3_objects(s3_key, stdin_key)

    def _decode_powershell_command(self, cmd_str: str) -> str:
        """
        Extract PowerShell script from -EncodedCommand if present.

        The wrapper expects PowerShell script code, not a command line with -EncodedCommand.
        If the command is a PowerShell command line with -EncodedCommand, decode it to get
        the actual script.

        WARNING: PowerShell -EncodedCommand uses UTF-16LE base64 encoding. We must decode
        this to plain text before uploading to S3 as UTF-8, otherwise the script will be
        double-encoded and fail to execute.

        :param cmd_str: The command string to check for -EncodedCommand
        :returns: Either the decoded PowerShell script or the original command
        """
        script_to_upload = cmd_str
        encoded_match = re.search(r'-EncodedCommand\s+["\']?([A-Za-z0-9+/=]+)["\']?', cmd_str)

        if encoded_match:
            self._verbosity_display(4, "EXEC_VIA_S3: Detected -EncodedCommand, decoding to extract PowerShell script")
            try:
                # WARNING: -EncodedCommand contains base64-encoded UTF-16LE PowerShell script
                # We MUST decode this to get the plain text script before uploading to S3
                # Changing this encoding will break execution of modules that use -EncodedCommand
                encoded_script = encoded_match.group(1)
                self._verbosity_display(5, f"EXEC_VIA_S3: Base64 length: {len(encoded_script)} chars")
                script_bytes = base64.b64decode(encoded_script)
                script_to_upload = script_bytes.decode("utf-16-le")
                self._verbosity_display(4, f"EXEC_VIA_S3: Decoded script length: {len(script_to_upload)} chars")
                self._verbosity_display(6, f"EXEC_VIA_S3: Decoded script:\n{script_to_upload}")
            except Exception as e:
                self._verbosity_display(
                    3, f"EXEC_VIA_S3: WARNING - Failed to decode -EncodedCommand, uploading as-is: {e}"
                )
                # Fall back to uploading the command line as-is
                script_to_upload = cmd_str

        return script_to_upload

    def _upload_script_to_s3(self, s3_key: str, script: str) -> str:
        """
        Upload PowerShell script to S3 and return presigned URL.

        WARNING: Scripts must be uploaded with UTF-8 encoding and ContentType.
        The PowerShell wrapper uses WebClient with UTF-8 encoding to download and
        decode the script. Changing this encoding will cause Unicode characters
        to be corrupted (garbled text, module failures, etc.).

        :param s3_key: S3 key for the script object
        :param script: PowerShell script content to upload
        :returns: Presigned URL for downloading the script
        """
        script_bytes = to_bytes(script, errors="surrogate_or_strict")
        self._verbosity_display(6, f"EXEC_VIA_S3: Script to upload (str): {repr(to_text(script)[:200])}")
        self._verbosity_display(6, f"EXEC_VIA_S3: Script to upload (bytes): {script_bytes[:200]}")

        # WARNING: Upload with UTF-8 ContentType - PowerShell wrapper expects UTF-8
        # Changing this will break Unicode support and cause encoding errors
        self._s3_client.put_object(
            Bucket=self._bucket_name,
            Key=s3_key,
            Body=script_bytes,
            ContentType="text/plain; charset=utf-8",
        )
        self._verbosity_display(4, f"EXEC_VIA_S3: Uploaded {len(script_bytes)} bytes to S3")

        # Generate presigned URL (1 hour expiry)
        presigned_url = self._s3_manager.get_url("get_object", self._bucket_name, s3_key, "GET")

        self._verbosity_display(4, f"EXEC_VIA_S3: Generated presigned URL (length: {len(presigned_url)})")
        self._verbosity_display(6, f"EXEC_VIA_S3: Presigned URL: {presigned_url}")

        return presigned_url

    def _upload_stdin_to_s3(self, stdin_key: str, in_data: bytes) -> str:
        """
        Upload stdin data to S3 and return presigned URL.

        :param stdin_key: S3 key for the stdin object
        :param in_data: Stdin data to upload
        :returns: Presigned URL for downloading the stdin data
        """
        self._verbosity_display(3, f"EXEC_VIA_S3: Uploading stdin data to s3://{self._bucket_name}/{stdin_key}")
        self._verbosity_display(4, f"EXEC_VIA_S3: Stdin data length: {len(in_data)} bytes")
        self._verbosity_display(6, f"EXEC_VIA_S3: Stdin data (bytes): {in_data[:200]}")
        self._verbosity_display(
            6,
            f"EXEC_VIA_S3: Stdin data (decoded): {repr(to_text(in_data, errors='surrogate_or_strict')[:200])}",
        )

        self._s3_client.put_object(
            Bucket=self._bucket_name,
            Key=stdin_key,
            Body=in_data,
            ContentType="text/plain; charset=utf-8",
        )

        stdin_url = self._s3_manager.get_url("get_object", self._bucket_name, stdin_key, "GET")
        self._verbosity_display(4, f"EXEC_VIA_S3: Generated stdin presigned URL (length: {len(stdin_url)})")
        self._verbosity_display(6, f"EXEC_VIA_S3: Stdin presigned URL: {stdin_url}")

        return stdin_url

    def _generate_wrapper_with_stdin(self, presigned_url: str, stdin_url: str, mark_begin: str, mark_end: str) -> str:
        """
        Generate PowerShell wrapper that downloads script and stdin from S3, then executes.

        The wrapper:
        1. Downloads script and stdin using WebClient (with UTF-8 encoding)
        2. Saves script to temp file
        3. Pipes stdin lines to PowerShell executing the script file
        4. Captures and outputs the exit code

        WARNING: ENCODING CHAIN:
        - S3 objects uploaded with ContentType="text/plain; charset=utf-8"
        - WebClient.Encoding=[System.Text.Encoding]::UTF8 to decode downloads
        - WriteAllText with UTF8 encoding to save script file
        All three steps MUST use UTF-8 or Unicode characters will be corrupted.

        :param presigned_url: Presigned URL for the script
        :param stdin_url: Presigned URL for stdin data
        :param mark_begin: Begin marker for output parsing
        :param mark_end: End marker for output parsing
        :returns: PowerShell wrapper command as string
        """
        # PowerShell: download script and stdin using WebClient (more reliable than Invoke-WebRequest)
        # See https://github.com/ansible-collections/amazon.aws/pull/2820
        # WARNING: Set WebClient encoding to UTF-8 to correctly decode S3 content
        # Save script to temp file and execute as separate process to handle exit codes correctly
        # Scriptblocks don't set $LASTEXITCODE properly when they contain exit statements
        wrapper = (
            f"try {{ "
            f"$wc=[System.Net.WebClient]::new() ; "
            # WARNING: WebClient.Encoding MUST be UTF8 to match S3 ContentType
            f"$wc.Encoding=[System.Text.Encoding]::UTF8 ; "
            f"$s=$wc.DownloadString('{presigned_url}') ; "
            f"$i=$wc.DownloadString('{stdin_url}') ; "
            f"$wc.Dispose() ; "
            f"$t=[System.IO.Path]::GetTempFileName() ; "
            f"$t=[System.IO.Path]::ChangeExtension($t,'.ps1') ; "
            # WARNING: WriteAllText MUST use UTF8 encoding to preserve Unicode
            f"[System.IO.File]::WriteAllText($t,$s,[System.Text.Encoding]::UTF8) ; "
            f"echo '{mark_begin}' ; "
            f'($i -split "`r?`n") | powershell -NoProfile -NonInteractive -ExecutionPolicy Bypass -File $t ; '
            f"$e=$LASTEXITCODE ; "
            f"Remove-Item -LiteralPath $t -Force -ErrorAction SilentlyContinue ; "
            f"echo '' ; echo $e ; echo '{mark_end}' "
            f"}} catch {{ "
            f"Write-Host '{mark_begin}' ; "
            f'Write-Host "S3_DOWNLOAD_ERROR: $_" ; '
            f"Write-Host '' ; "
            f"Write-Host {S3_DOWNLOAD_ERROR_EXIT_CODE} ; "
            f"Write-Host '{mark_end}' "
            f"}}"
        )
        return wrapper

    def _generate_wrapper_without_stdin(self, presigned_url: str, mark_begin: str, mark_end: str) -> str:
        """
        Generate PowerShell wrapper that downloads script from S3 and executes (no stdin).

        The wrapper:
        1. Downloads script using WebClient (with UTF-8 encoding)
        2. Saves script to temp file
        3. Executes the script file in a new PowerShell process
        4. Captures and outputs the exit code

        WARNING: ENCODING CHAIN:
        - S3 objects uploaded with ContentType="text/plain; charset=utf-8"
        - WebClient.Encoding=[System.Text.Encoding]::UTF8 to decode downloads
        - WriteAllText with UTF8 encoding to save script file
        All three steps MUST use UTF-8 or Unicode characters will be corrupted.

        :param presigned_url: Presigned URL for the script
        :param mark_begin: Begin marker for output parsing
        :param mark_end: End marker for output parsing
        :returns: PowerShell wrapper command as string
        """
        # No stdin: download script using WebClient and execute
        # See https://github.com/ansible-collections/amazon.aws/pull/2820
        # WARNING: Set WebClient encoding to UTF-8 to correctly decode S3 content
        # Save script to temp file and execute as separate process to handle exit codes correctly
        # Scriptblocks don't set $LASTEXITCODE properly when they contain exit statements
        wrapper = (
            f"try {{ "
            f"$wc=[System.Net.WebClient]::new() ; "
            # WARNING: WebClient.Encoding MUST be UTF8 to match S3 ContentType
            f"$wc.Encoding=[System.Text.Encoding]::UTF8 ; "
            f"$s=$wc.DownloadString('{presigned_url}') ; "
            f"$wc.Dispose() ; "
            f"$t=[System.IO.Path]::GetTempFileName() ; "
            f"$t=[System.IO.Path]::ChangeExtension($t,'.ps1') ; "
            # WARNING: WriteAllText MUST use UTF8 encoding to preserve Unicode
            f"[System.IO.File]::WriteAllText($t,$s,[System.Text.Encoding]::UTF8) ; "
            f"echo '{mark_begin}' ; "
            f"powershell -NoProfile -NonInteractive -ExecutionPolicy Bypass -File $t ; "
            f"$e=$LASTEXITCODE ; "
            f"Remove-Item -LiteralPath $t -Force -ErrorAction SilentlyContinue ; "
            f"echo '' ; echo $e ; echo '{mark_end}' "
            f"}} catch {{ "
            f"Write-Host '{mark_begin}' ; "
            f'Write-Host "S3_DOWNLOAD_ERROR: $_" ; '
            f"Write-Host '' ; "
            f"Write-Host {S3_DOWNLOAD_ERROR_EXIT_CODE} ; "
            f"Write-Host '{mark_end}' "
            f"}}"
        )
        return wrapper

    def _execute_wrapper(self, wrapper: str, mark_begin: str, mark_end: str) -> tuple[int, str, str]:
        """
        Send wrapper command to session and execute via normal communicate flow.

        :param wrapper: PowerShell wrapper command to execute
        :param mark_begin: Begin marker for output parsing
        :param mark_end: End marker for output parsing
        :returns: Tuple of (exit_code, stdout, stderr)
        """
        self._verbosity_display(3, f"EXEC_VIA_S3: Wrapper command length: {len(wrapper)} bytes")
        self._verbosity_display(5, f"EXEC_VIA_S3: Wrapper command:\n{wrapper}")

        # Send wrapper command via session (tiny, won't timeout from echo)
        wrapper_bytes = to_bytes(wrapper + "\n", errors="surrogate_or_strict")
        self._verbosity_display(3, f"EXEC_VIA_S3: Sending wrapper command to session ({len(wrapper_bytes)} bytes)")
        self._session_manager.stdin_write(wrapper_bytes)

        # Execute via normal communicate flow (no stdin data since wrapper handles it)
        return self._exec_communicate(wrapper, None, mark_begin, mark_end)

    def _cleanup_s3_objects(self, s3_key: str, stdin_key: Optional[str]) -> None:
        """
        Clean up temporary S3 objects created for command execution.

        :param s3_key: S3 key for the script object
        :param stdin_key: Optional S3 key for stdin data object
        """
        try:
            self._verbosity_display(4, f"EXEC_VIA_S3: Cleaning up S3 object: {s3_key}")
            self._s3_client.delete_object(Bucket=self._bucket_name, Key=s3_key)
            if stdin_key is not None:
                self._verbosity_display(4, f"EXEC_VIA_S3: Cleaning up S3 stdin object: {stdin_key}")
                self._s3_client.delete_object(Bucket=self._bucket_name, Key=stdin_key)
        except Exception as e:
            self._verbosity_display(3, f"EXEC_VIA_S3: Failed to clean up S3 objects: {e}")



# -*- coding: utf-8 -*-

# Copyright: Contributors to the Ansible project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

import base64
import re
import time

try:
    # ansible-core >= 2.21
    from ansible._internal._powershell import _script

    HAS_MODERN_PWSH = True
except ImportError:
    # ansible-core < 2.21 fallback
    HAS_MODERN_PWSH = False
    _script = None  # For type checking

from ansible.module_utils.common.text.converters import to_bytes
from ansible.module_utils.common.text.converters import to_text

from ansible_collections.amazon.aws.plugins.plugin_utils.ssm.common import generate_mark


class TerminalManager:
    def __init__(self, connection):
        """
        Initialise the TerminalManager.

        :param connection: The AWS SSM connection object
        """
        self.connection = connection

    def prestage_windows_command_executor(self) -> None:
        """Pre-stage a PowerShell function to execute commands from stdin.

        This avoids PTY echo issues by reading the command itself from stdin
        rather than sending it through the terminal. The function also supports
        reading additional stdin data for the command if needed.

        Uses extremely compressed syntax to minimize PTY echo delays.
        Does NOT wait for confirmation - the prestage command will be echoed
        slowly by the PTY, but we proceed immediately. If the function fails
        to load, the first actual command will fail.
        """
        # Compressed function that:
        # 1. Reads command from stdin (until EOF delimiter \0\0\0\0)
        # 2. Optionally reads stdin data for command (until second EOF delimiter)
        # 3. Executes command, piping stdin if provided
        # 4. Returns output wrapped in markers
        # Variable names: $b=BeginMark, $e=EndMark, $h=HasStdin, $c=cmd, $l=line, $t=cmdText, $d=stdinData
        command = (
            "function i($b,$e,$h){"
            "$c=@();"
            'while(($l=[Console]::In.ReadLine())-and($l-ne"`0`0`0`0")){$c+=$l};'
            '$t=$c-join"`n";'
            "echo $b;"
            "if($h){"
            "$d=@();"
            'while(($l=[Console]::In.ReadLine())-and($l-ne"`0`0`0`0")){$d+=$l};'
            "$d|iex $t"
            "}else{"
            "iex $t"
            "};"
            "echo '';"
            "echo $LASTEXITCODE;"
            "echo $e"
            "}\n"
        )

        prestage_cmd = to_bytes(command, errors="surrogate_or_strict")
        self.connection.verbosity_display(3, "PRESTAGE CMD EXEC: Sending compressed stdin helper function 'i'")
        self.connection.verbosity_display(4, f"PRESTAGE CMD EXEC: Command length: {len(prestage_cmd)} bytes")
        self.connection.verbosity_display(5, f"PRESTAGE CMD EXEC: Command:\n{to_text(prestage_cmd)}")
        self.connection.verbosity_display(
            3,
            "PRESTAGE CMD EXEC: Not waiting for confirmation (PTY echo too slow) - will fail on first command if load failed",
        )

        # Send command without waiting for completion
        # The PTY will echo this slowly, but PowerShell will eventually execute it
        # If it fails to load, the first actual command will error
        self.connection.session_manager.stdin_write(prestage_cmd)

        # Give PowerShell a moment to process the function definition
        # This is a compromise - we can't wait for confirmation due to PTY echo,
        # but we also can't proceed instantly or the function won't be ready
        time.sleep(2)
        self.connection.verbosity_display(3, "PRESTAGE CMD EXEC: Proceeding without confirmation")

    def prepare_windows(self) -> None:
        self.disable_windows_echo_command()
        self.disable_windows_prompt_command()
        self.prestage_windows_command_executor()
        self.connection.verbosity_display(3, "PRE (windows) Terminal configured")

    def prepare_default(self) -> None:
        # Ensure SSM Session has started
        self.ensure_ssm_session_has_started()
        # Disable echo command
        self.disable_echo_command()
        # Disable prompt command
        self.disable_prompt_command()
        self.connection.verbosity_display(3, "PRE (default) Terminal configured")

    def prepare_terminal(self) -> None:
        """perform any one-time terminal settings"""
        if self.connection.is_windows:
            self.prepare_windows()
        else:
            self.prepare_default()

    def _encode_powershell_command(self, script: str, disable_input: bool = False) -> str:
        """Encode a PowerShell script using the appropriate method for the Ansible version.

        :param script: The PowerShell script to encode.
        :param disable_input: Whether to disable stdin (only supported in modern versions).
        :returns: The encoded PowerShell command line.
        """
        self.connection.verbosity_display(
            6, f"ENCODE_PWSH: Script length: {len(script)} chars, disable_input={disable_input}"
        )
        if HAS_MODERN_PWSH:
            # Use modern API (ansible-core >= 2.21)
            self.connection.verbosity_display(5, "ENCODE_PWSH: Using modern _script.get_pwsh_encoded_cmdline() API")
            self.connection.verbosity_display(6, f"ENCODE_PWSH: Input script:\n{script}")
            cmd_parts = _script.get_pwsh_encoded_cmdline(
                script,
                override_execution_policy=True,
                disable_input=disable_input,
            )
            result = " ".join(cmd_parts)
            self.connection.verbosity_display(5, f"ENCODE_PWSH: Result length: {len(result)} chars")
            self.connection.verbosity_display(6, f"ENCODE_PWSH: Result command:\n{result}")
            return result
        else:
            # Fallback for ansible-core < 2.21
            self.connection.verbosity_display(5, "ENCODE_PWSH: Using legacy base64 encoding (ansible-core < 2.21)")
            self.connection.verbosity_display(6, f"ENCODE_PWSH: Input script:\n{script}")
            # Encode script to base64 UTF-16LE
            script_bytes = to_bytes(script, encoding="utf-16-le")
            encoded_script = base64.b64encode(script_bytes).decode("ascii")
            self.connection.verbosity_display(6, f"ENCODE_PWSH: Base64 length: {len(encoded_script)} chars")

            # Build command line (mimics _common_args from powershell.py)
            # Note: Legacy method doesn't support disable_input parameter
            if disable_input:
                self.connection.verbosity_display(
                    5, "ENCODE_PWSH: WARNING - disable_input not supported in legacy mode, ignoring"
                )
            result = (
                f"powershell -NoProfile -NonInteractive -ExecutionPolicy Unrestricted -EncodedCommand {encoded_script}"
            )
            self.connection.verbosity_display(5, f"ENCODE_PWSH: Result length: {len(result)} chars")
            self.connection.verbosity_display(6, f"ENCODE_PWSH: Result command:\n{result}")
            return result

    def wrap_command(self, cmd: str, mark_begin: str, mark_end: str, has_stdin: bool = False) -> tuple[str, str]:
        """Wrap command so stdout and status can be extracted.

        For Windows: Returns a trigger command that invokes the pre-staged executor.
                     The actual command is returned separately to be sent via stdin.
        For Linux: Returns the wrapped command inline (original behavior).

        :param cmd: The command to wrap.
        :param mark_begin: The begin marker.
        :param mark_end: The end marker.
        :param has_stdin: Whether stdin data will be sent to this command.
        :returns: Tuple of (trigger_command, actual_command_for_stdin).
                  For Linux, actual_command_for_stdin will be empty string.
        """

        if self.connection.is_windows:
            # Windows: Use pre-staged helper function 'i'
            # The trigger is just the function call - command is sent via stdin
            has_stdin_str = "$true" if has_stdin else "$false"
            trigger_cmd = f"i '{mark_begin}' '{mark_end}' {has_stdin_str}\n"

            self.connection.verbosity_display(4, f"WRAP_COMMAND: Windows mode (has_stdin={has_stdin})")
            self.connection.verbosity_display(5, f"WRAP_COMMAND: Trigger command: {trigger_cmd.strip()}")
            self.connection.verbosity_display(5, f"WRAP_COMMAND: Command to send via stdin ({len(cmd)} bytes)")
            self.connection.verbosity_display(6, f"WRAP_COMMAND: Full command for stdin:\n{to_text(cmd)}")

            return (trigger_cmd, cmd)
        else:
            # Linux/Unix: use existing inline approach
            self.connection.verbosity_display(4, "WRAP_COMMAND: Linux/Unix mode - wrapping command")
            self.connection.verbosity_display(5, f"WRAP_COMMAND: Original command: {to_text(cmd)[:200]}...")
            self.connection.verbosity_display(5, f"WRAP_COMMAND: Original command length: {len(cmd)} bytes")
            self.connection.verbosity_display(6, f"WRAP_COMMAND: Full original command:\n{to_text(cmd)}")
            wrapped_cmd = (
                f"printf '%s\\n' '{mark_begin}' ;\n"
                f"echo | {cmd} ;\n"
                f"printf '\\n%s\\n%s\\n' \"$?\" '{mark_end}' ;\n"
            )  # fmt: skip
            self.connection.verbosity_display(5, f"WRAP_COMMAND: Wrapped command length: {len(wrapped_cmd)} bytes")
            self.connection.verbosity_display(5, f"WRAP_COMMAND: Begin marker: {mark_begin}")
            self.connection.verbosity_display(5, f"WRAP_COMMAND: End marker: {mark_end}")
            self.connection.verbosity_display(6, f"WRAP_COMMAND: Full wrapped command:\n{to_text(wrapped_cmd)}")

            return (wrapped_cmd, "")

    def disable_echo_command(self) -> None:
        """Disable terminal echo on Linux/Unix using stty.

        This prevents commands from being echoed back, which improves output
        parsing and prevents sensitive data from appearing in logs.
        """
        disable_echo_cmd = to_bytes("stty -echo\n", errors="surrogate_or_strict")

        # Send command
        self.connection.verbosity_display(3, "DISABLE ECHO: Disabling terminal echo using stty")
        self.connection.verbosity_display(4, f"DISABLE ECHO Command: {to_text(disable_echo_cmd).strip()}")
        self.connection.verbosity_display(5, f"DISABLE ECHO Command bytes: {len(disable_echo_cmd)} bytes")
        self.connection.session_manager.stdin_write(disable_echo_cmd)

        self.connection.session_manager.wait_for_match(
            label="DISABLE ECHO", cmd=disable_echo_cmd, match="stty -echo", is_windows=self.connection.is_windows
        )
        self.connection.verbosity_display(3, "DISABLE ECHO: Terminal echo disabled")

    def disable_windows_echo_command(self) -> None:
        """Disable command echo in PowerShell.

        Disables both PSReadLine (syntax highlighting) and Windows Console echo.
        PSReadLine provides interactive features including command echo with syntax
        highlighting. The Windows Console also echoes input via ENABLE_ECHO_INPUT flag.
        """
        end_mark = generate_mark()
        # Remove PSReadLine and disable console echo via Windows API
        # This is equivalent to 'stty -echo' on Linux
        command = (
            "Remove-Module PSReadLine -ErrorAction SilentlyContinue ; "
            # Define P/Invoke for kernel32 console functions
            "Add-Type -TypeDefinition @'\n"
            "using System;\n"
            "using System.Runtime.InteropServices;\n"
            "public class ConsoleHelper {\n"
            '    [DllImport("kernel32.dll", SetLastError = true)]\n'
            "    public static extern IntPtr GetStdHandle(int nStdHandle);\n"
            '    [DllImport("kernel32.dll", SetLastError = true)]\n'
            "    public static extern bool GetConsoleMode(IntPtr hConsoleHandle, out uint lpMode);\n"
            '    [DllImport("kernel32.dll", SetLastError = true)]\n'
            "    public static extern bool SetConsoleMode(IntPtr hConsoleHandle, uint dwMode);\n"
            "    public const int STD_INPUT_HANDLE = -10;\n"
            "    public const uint ENABLE_ECHO_INPUT = 0x0004;\n"
            "}\n"
            "'@ -ErrorAction SilentlyContinue ; "
            # Disable echo on stdin
            "$h = [ConsoleHelper]::GetStdHandle([ConsoleHelper]::STD_INPUT_HANDLE) ; "
            "$mode = [uint32]0 ; "
            "[void][ConsoleHelper]::GetConsoleMode($h, [ref]$mode) ; "
            "$mode = $mode -band (-bnot [ConsoleHelper]::ENABLE_ECHO_INPUT) ; "
            "[void][ConsoleHelper]::SetConsoleMode($h, $mode) ; "
            f"echo {end_mark}\n"
        )
        disable_echo_cmd = to_bytes(command, errors="surrogate_or_strict")
        end_marker_rex = f"^{re.escape(end_mark)}$"
        self.connection.verbosity_display(3, "DISABLE WIN ECHO: Disabling PSReadLine and Windows Console echo")
        self.connection.verbosity_display(4, f"DISABLE WIN ECHO End marker text: {end_mark}")
        self.connection.verbosity_display(4, f"DISABLE WIN ECHO End marker RegEx: \n{end_marker_rex}")
        disable_echo_reply = re.compile(end_marker_rex, re.MULTILINE)

        # Send command
        self.connection.verbosity_display(4, f"DISABLE WIN ECHO Command text: \n{to_text(disable_echo_cmd)}")
        self.connection.verbosity_display(4, f"DISABLE WIN ECHO Sending {len(disable_echo_cmd)} bytes")
        self.connection.session_manager.stdin_write(disable_echo_cmd)

        self.connection.session_manager.wait_for_match(
            label="DISABLE WIN ECHO", cmd=disable_echo_cmd, match=disable_echo_reply.search, is_windows=True
        )
        self.connection.verbosity_display(3, "DISABLE WIN ECHO: PSReadLine removed and Console echo disabled")

    def disable_windows_prompt_command(self) -> None:
        """Disable prompt in PowerShell by setting it to empty string.

        This mirrors the Linux behavior where PS1='' sets an empty prompt.
        """
        end_mark = generate_mark()
        command = f'function prompt {{ "" }} ; echo {end_mark}\n'
        disable_prompt_cmd = to_bytes(command, errors="surrogate_or_strict")
        end_marker_rex = f"^{re.escape(end_mark)}$"
        self.connection.verbosity_display(4, f"DISABLE WIN PROMPT End marker text: {end_mark}")
        self.connection.verbosity_display(4, f"DISABLE WIN PROMPT End marker RegEx: \n{end_marker_rex}")
        disable_prompt_reply = re.compile(end_marker_rex, re.MULTILINE)

        # Send command
        self.connection.verbosity_display(4, f"DISABLE WIN PROMPT Command text: \n{to_text(disable_prompt_cmd)}")
        self.connection.verbosity_display(4, f"DISABLE WIN PROMPT Sending {len(disable_prompt_cmd)} bytes")
        self.connection.session_manager.stdin_write(disable_prompt_cmd)

        self.connection.session_manager.wait_for_match(
            label="DISABLE WIN PROMPT", cmd=disable_prompt_cmd, match=disable_prompt_reply.search, is_windows=True
        )

    def disable_prompt_command(self) -> None:
        """Disable prompt on Linux/Unix by setting PS1 to empty string.

        This mirrors the Windows behavior of setting an empty prompt function.
        Also disables bracketed paste mode which can interfere with output parsing.
        """
        end_mark = generate_mark()
        command = (
            "PS1='' ; "
            "bind 'set enable-bracketed-paste off'; "
            f"printf '\\n%s\\n' '{end_mark}'\n"
        )  # fmt: skip
        disable_prompt_cmd = to_bytes(command, errors="surrogate_or_strict")
        disable_prompt_reply = re.compile(r"\r\r\n" + re.escape(end_mark) + r"\r\r\n", re.MULTILINE)

        # Send command
        self.connection.verbosity_display(
            3, "DISABLE PROMPT: Setting empty prompt (PS1='') and disabling bracketed paste"
        )
        self.connection.verbosity_display(4, f"DISABLE PROMPT End marker: {end_mark}")
        self.connection.verbosity_display(5, f"DISABLE PROMPT Command: {to_text(disable_prompt_cmd).strip()}")
        self.connection.session_manager.stdin_write(disable_prompt_cmd)

        self.connection.session_manager.wait_for_match(
            label="DISABLE PROMPT", cmd=disable_prompt_cmd, match=disable_prompt_reply.search, is_windows=False
        )
        self.connection.verbosity_display(3, "DISABLE PROMPT: Prompt disabled")

    def ensure_ssm_session_has_started(self) -> None:
        """Ensure the SSM session has started on the host.

        Polls stdout until we match 'Starting session with SessionId' which
        indicates the session-manager-plugin has successfully connected.
        """
        self.connection.verbosity_display(3, "START SSM SESSION: Waiting for session to start")
        self.connection.verbosity_display(4, "START SSM SESSION: Looking for 'Starting session with SessionId' message")
        self.connection.session_manager.wait_for_match(
            label="START SSM SESSION",
            cmd="start_session",
            match="Starting session with SessionId",
            is_windows=self.connection.is_windows,
        )
        self.connection.verbosity_display(3, "START SSM SESSION: Session started successfully")
